from pathlib import Path
from bridgeobjects import load_pbn


TEST_PATH = Path('tests', 'test_data')


def get_boards(file_name: str) -> dict[str, object]:
    event_number = 0
    board_path = Path(TEST_PATH, file_name)
    raw_boards = load_pbn(board_path)[event_number].boards
    boards = {}
    for board in raw_boards:
        boards[board.identifier] = board
        played_cards = []
        for trick in board.tricks:
            played_cards += trick.cards
        for hand in board.hands.values():
            hand.unplayed_cards = [card for card in hand.cards if card not in played_cards]
    return boards




from bridgeobjects import Card, Suit


SUIT_SELECTION_REASON = {
    '000': 'No reason applied',
    '001': '5+ card suit with honour',
    '002': 'Best suit sequence',
    '003': 'Best long suit',
    '004': 'Long suit with most honours',
    '005': 'Only four card suit',
    '006': 'Best of 4 card suits',
    '007': 'Long_suit with no ace',
    '008': 'Strongest long suit',
    '009': '4+ card suit with touching honours',
    '010': 'Best 4 card suit with no split honours',
    '011': 'Best 4 card suit with near sequence',
    '012': 'Best 4 card suit with interior sequence',
    '013': 'Suit bid by partner',
    '014': 'Only non-deprecated suit',
    '015': 'Non-deprecated suit - 2 honours',
    '016': 'Non-deprecated suit - doubleton, no honour',
    '017': 'Non-deprecated suit - no honour',
    '018': 'Longest non-deprecated suit',
    '019': 'Unbid major',
    '020': "Partner's suit",
    '021': '5+ card suit with entries',
    '022': '3+ card suit with no honour',
    '023': 'Single honour suit headed by K',
    '024': 'Single honour suit headed by Q',
    '025': 'Single honour suit headed by J',
    '026': "Probable shortage in partner's hand",
    '027': 'Touching honours',
    '028': 'Trumps',
    '029': 'Long suit',
    '030': 'Defending slam - try to establish second trick',
    '031': 'Defending slam - partner indicated suit',
    '032': 'Defending slam - partner indicated suit: unusual double',
}

CARD_SELECTION_REASON = {
    '000': 'No reason applied',
    '001': 'Top of doubleton',
    '002': 'K from AK doubleton',
    '003': 'Top of touching honours',
    '004': 'Fourth highest',
    '005': 'Middle with no honour',
    '006': 'Fourth highest from a single honour',
    '007': 'Bottom of triplet with an honour',
    '008': 'Top of touching honours in triplet',
    '009': 'Top of internal sequence',
    '010': 'Triplet with two honours, third card',
    '011': "Bottom from QTx",
    '013': 'A from suit headed by an A',
    '014': 'Second highest from 4 rags',
    '015': 'Singleton',
    '020': 'Top of solid sequence',
    '021': 'Top of near sequence',
    '022': 'Top of internal sequence',
    '023': 'Top of touching honours',
}


class SelectedSuit(Suit):
    """An extension of the Suit class that contains a
    reason code for suit_selection."""

    def __init__(self, name=None, selection_reason='000', *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self._selection_reason = selection_reason

    def __str__(self):
        """Return the str value."""
        return (f'Suit("{self.name}"), '
                f'{SUIT_SELECTION_REASON[self._selection_reason]}')

    @property
    def selection_reason(self):
        """Return the value of the property selection_reason."""
        return self._selection_reason

    @selection_reason.setter
    def selection_reason(self, value):
        """Set the value of the property selection_reason."""
        self._selection_reason = value


class SelectedCard(Card):
    """An extension of the Card class that contains a reason code for
    suit_selection."""

    def __init__(self, name, selection_reason='000', *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.selection_reason = selection_reason

    def __str__(self):
        """Return the str value."""
        selection_reason = CARD_SELECTION_REASON[self.selection_reason]
        return (f'Card("{self.name}"), '
                f'{self.selection_reason}, {selection_reason}')



"""Fourth seat card play for defender."""

import random

from bridgeobjects.src.constants import CARD_VALUES
from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import SUITS, Card, Suit
from bfgdealer import Trick
import bfgcardplay.src.global_variables as global_vars

from .player import Player
from .utilities import (get_list_of_best_scores, get_suit_strength,
                        trick_card_values)
from .fourth_seat import FourthSeat
from .defender_play import get_hilo_signal_card, signal_card, surplus_card, best_discard

MODULE_COLOUR = 'cyan'
PLAYER_INDEX = 3


class FourthSeatDefender(FourthSeat):
    def __init__(self, player: Player):
        super().__init__(player)
        self.player = player

    def selected_card(self) -> Card:
        """Return the card if the third seat."""
        player = self.player
        trick = player.board.tricks[-1]

        cards = player.cards_for_trick_suit(trick)

        # Void
        if not cards:
            return self._select_card_if_void(player, trick)

        # Singleton
        if len(cards) == 1:
            return log(inspect.stack(), cards[0])

        # play low if partner is winning trick
        if self._second_player_winning_trick(cards, trick, player.trump_suit):
            return log(inspect.stack(), cards[-1])

        # win trick if possible
        winning_card = self._winning_card(trick)
        unplayed_cards = player.total_unplayed_cards[trick.suit.name]
        if winning_card:
            value = winning_card.value
            play_winner = False
            while value > 1:
                value -= 1
                card = Card(CARD_VALUES[value], trick.suit.name)
                if (card in unplayed_cards and
                        card not in player.dummys_unplayed_cards[trick.suit.name]):
                    play_winner = True
                    break
            if play_winner:
                return log(inspect.stack(), winning_card)

        # Play low if higher card is winner after trick played
        # losing_trick = (trick.cards[0].value > trick.cards[1].value or
        #                     trick.cards[2].value > trick.cards[1].value)
        # if len(cards) == 2 and losing_trick:
        #     winner = True
        #     for card in unplayed_cards:
        #         if card.value > cards[0].value:
        #             winner = False
        #     if winner:
        #         return log(inspect.stack(), cards[1])

        # Signal even/odd
        return get_hilo_signal_card(self)

    def _select_card_if_void(self, player: Player, trick: Trick) -> Card:
        """Return card if cannot follow suit."""
        manager = global_vars.manager

        # Trump if appropriate
        if player.trump_suit:
            values = trick_card_values(trick, player.trump_suit)
            if player.trump_cards:
                if values[0] > values[1] or values[2] > values[1]:
                    for card in player.trump_cards[::-1]:
                        if card.value + 13 > values[0] and card.value + 13 > values[2]:
                            return log(inspect.stack(), card)

        # Do not signal with winners
        # for suit in SUITS:
        #     for card in player.unplayed_cards[suit]:
        #         if not player.is_master_card(card):
        #             return log(inspect.stack(), card)

        # Signal best suit
        best_suit = self._best_suit()
        card = signal_card(player, manager, best_suit)
        if card:
            if not player.is_master_card(card):
                return log(inspect.stack(), card)

        # Discard if more cards than the opposition
        card = surplus_card(player)
        if card:
            return log(inspect.stack(), card)

        # Best discard
        return log(inspect.stack(), best_discard(player))

    def _best_suit(self) -> Suit:
        """Select suit for signal."""
        # TODO handle no points and equal suits
        player = self.player
        cards = player.hand_cards.list
        suit_points = get_suit_strength(cards)
        max_points = 0
        best_suit = self._strongest_suit()
        return best_suit
        for suit in SUITS:
            if suit != player.trump_suit.name:
                hcp = suit_points[suit]
                if hcp > max_points:
                    max_points = hcp
                    best_suit = suit
        if not best_suit:
            for suit in SUITS:
                if suit != player.trump_suit.name:
                    hcp = suit_points[suit]
                    if hcp > max_points:
                        max_points = hcp
                        best_suit = suit
        if not best_suit:
            return player.trump_suit
        return Suit(best_suit)

    def _strongest_suit(self) -> Suit | None:
        """Return the strongest_suit."""
        player = self.player
        suits = {suit: 0 for suit in SUITS}
        for suit in SUITS:
            cards = player.unplayed_cards[suit]
            for card in cards:
                suits[suit] += card.value
        if player.trump_suit:
            suits[player.trump_suit.name] = 0
        best_suits = get_list_of_best_scores(suits)
        # if not best_suits:
        #     return player.trump_suit
        return Suit(random.choice(best_suits))



"""Test opening lead cards."""
from bridgeobjects import Hand, Auction, Contract, Suit, Board

from ..src.opening_lead_suit import opening_lead_suit


def _get_board(dealer, hand, bids):
    """Return a board based on parameters."""
    board = Board()
    board.dealer = dealer
    board.auction = Auction(bids, board.dealer)
    board.contract = Contract(auction=board.auction)
    board.hands[board.contract.leader] = hand
    return board


def test_missing_partners_suit():
    """Test that a 4 card suit is selected against NT contract. Partner not bid."""
    hand = Hand('AT98642.754..874')
    bids = ['1H', '2D', '3NT', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit != Suit('D')


def test_four_card_suit():
    """Test that a 4 card suit is selected against NT contract. Partner not bid."""
    hand = Hand('8642.754.A94.987')
    bids = ['1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '005'


def test_best_four_card_suit():
    """Test that the 4 card long suit is selected against NT contract. Partner not bid."""
    hand = Hand('AJ83.754.AQ94.98')
    bids = ['P', '1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '006'


def test_long_suit():
    """Test that the long suit is selected against NT contract. Partner not bid."""
    hand = Hand('A9832.75.AK94.98')
    bids = ['1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('E', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '003'


def test_best_longest_suit():
    """Test that the best longest suit is selected against NT contract. Partner not bid."""
    hand = Hand('A9832.7.AK954.98')
    bids = ['1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('E', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '009'


def test_klinger_3_B_1_i():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('752.87.AJT94.985')
    bids = ['P', '1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '001'


def test_klinger_3_B_ii_1_a():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('752.87.AJT94.985')
    bids = ['P', '1D', 'P', '1NT', 'P', '3D', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '019'


def test_klinger_3_B_ii_1_b():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('752.87.AJT94.985')
    bids = ['P', 'P', '1H', '1NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '020'


def test_klinger_3_B_2_i():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('QJ982.Q4.JT32.98')
    bids = ['1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '001'


def test_klinger_3_B_2_ii_a():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('QJ982.Q4.JT32.98')
    bids = ['1S', 'P', '2H', 'P', '2NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '019'


def test_klinger_3_B_2_ii_b():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('QJ982.Q4.JT32.98')
    bids = ['1NT', 'P', '2C', 'P', '2H', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '019'


def test_klinger_3_B_3_i():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('632.K9753.QT7.J8')
    bids = ['P', '1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '003'


def test_klinger_3_B_3_ii_a():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('632.K9753.QT7.J8')
    bids = ['1NT', 'P', '3H', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '019'


def test_klinger_3_B_3_ii_b():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('632.K9753.QT7.J8')
    bids = ['2NT', 'P', '3S', 'P', '4S', 'P', '4NT', 'P', '5H', 'P', '6NT', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '022'


def test_klinger_3_B_4_i():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('QT982.J83.J732.4')
    bids = ['P', '1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '001'


def test_klinger_3_B_4_ii_a():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('QT982.J83.J732.4')
    bids = ['P', '1S', 'P', '2D', 'P', '3D', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '019'


# def test_klinger_3_B_4_ii_a(): # Too esoteric
    # """Klinger exercise. Better Card play p15."""
#     hand = Hand('QT982.J83.J732.4')
#     seat = 'W'
#     dealer = 'W'
#     auction = Auction(['P', '1C', 'P', '1NT', 'P', '2NT', 'P', '3NT', 'P', 'P', 'D', 'P', 'P', 'P'], dealer)
#     suit = opening_lead_suit(hand, seat, dealer, auction)
#     assert suit == Suit('C')
#     assert suit.selection_reason == '019'


def test_klinger_3_B_5_i():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('JT986.7.AQT.7432')
    bids = ['P', '1NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')


def test_klinger_3_B_5_iA():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('JT986.7.AQT.7432')
    bids = ['P', '1D', '1H', '2D', 'P', '2H', 'P', '2NT', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '021'


def test_klinger_3_B_5_ii():
    """Klinger exercise. Better Card play p15."""
    hand = Hand('JT986.7.AQT.7432')
    bids = ['P', '3C', 'P', '3NT', 'P', 'P', 'P']
    board = _get_board('W', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '021'


def test_klinger_4_1_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('AJ2.J7.9743.A852')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '014'


def test_klinger_4_1_ii_a():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('AJ2.J7.9743.A852')
    dealer = 'S'
    bids = ['1S', 'P', '2D', 'P', '2S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '016'


def test_klinger_4_1_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('AJ2.J7.9743.A852')
    dealer = 'W'
    bids = ['P', '1NT', '2H', '4S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '020'


def test_klinger_4_2_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('K2.J843.K83.A964')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '023'


def test_klinger_4_2_ii_a():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('K2.J843.K83.A964')
    dealer = 'S'
    bids = ['1S', 'P', '2C', 'P', '3C', 'P', '4S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('C')
    assert suit.selection_reason == '026'


def test_klinger_4_2_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('K2.J843.K83.A964')
    dealer = 'S'
    bids = ['1H', 'P', '2C', 'P', '2S', 'P', '3H', 'P', '4H', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '014'


def test_klinger_4_3_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('83.9.KJ752.Q9843')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '008'


# def test_klinger_4_3_ii_a():
    # """Klinger exercise. Better Card play p24."""
#     hand = Hand('83.9.KJ752.Q9843')
#     dealer = 'S'
#     bids = ['1C', 'P', '1S', 'P', '2C', 'P', 'P', 'P']
#     board = _get_board(dealer, hand, bids)
#     suit = opening_lead_suit(board)

#     assert suit == Suit('D')  # not sure of reason for D!!!
#     assert suit.selection_reason == '026'


def test_klinger_4_3_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('83.9.KJ752.Q9843')
    dealer = 'S'
    bids = ['1H', 'P', '2C', 'P', '3C', 'P', '4H', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('C')
    assert suit.selection_reason == '026'


def test_klinger_4_4_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('AQ8.832.QJ72.843')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '027'


def test_klinger_4_4_ii_a():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('AQ8.832.QJ72.843')
    dealer = 'S'
    bids = ['1S', 'P', '1NT', 'P', '2H', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '028'


def test_klinger_4_4_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('AQ8.832.QJ72.843')
    dealer = 'S'
    bids = ['1C', 'P', '1D', 'P', '1H', 'P', '2H', 'P', '3H', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '028'


def test_klinger_4_5_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('KJ6.A975.2.KQ754')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('C')
    assert suit.selection_reason == '027'


def test_klinger_4_5_ii_a():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('KJ6.A975.2.KQ754')
    dealer = 'S'
    bids = ['1D', 'D', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '028'


def test_klinger_4_5_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('KJ6.A975.2.KQ754')
    dealer = 'S'
    bids = ['3H', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('C')
    assert suit.selection_reason == '027'


def test_klinger_4_6_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('98.AJ64.K763.873')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '028'


def test_klinger_4_6_ii_a():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('98.AJ64.K763.873')
    dealer = 'W'
    bids = ['P', '1C', 'P', '1S', 'P', '3C', 'P', '3S', '4S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '029'


def test_klinger_4_6_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('98.AJ64.K763.873')
    dealer = 'S'
    bids = ['1S', 'P', '1NT', 'P', '2H', 'P', '2S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('S')
    assert suit.selection_reason == '028'


def test_klinger_4_7_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('A8.7.K872.976432')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '008'


def test_klinger_4_7_ii_a():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('A8.7.K872.976432')
    dealer = 'S'
    bids = ['1S', 'P', '3S', 'P', '4NT', 'P', '5H', '6S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '030'


def test_klinger_4_7_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('A8.7.K872.976432')
    dealer = 'S'
    bids = ['1S', 'P', '3D', 'P', '4D', 'P', '4NT', 'P', '5H', '6S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '030'


def test_klinger_4_8_i():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('K2.86.QJT7.98642')
    bids = ['1S', 'P', '3S', 'P', '4S', 'P', 'P', 'P']
    board = _get_board('S', hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('D')
    assert suit.selection_reason == '002'


def test_klinger_4_8_ii_a():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('K2.86.QJT7.98642')
    dealer = 'N'
    bids = ['1C', 'P', '2S', 'P', '3S', 'P', '4NT', 'P', '5H', 'D', '6S', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('H')
    assert suit.selection_reason == '031'


def test_klinger_4_8_ii_b():
    """Klinger exercise. Better Card play p24."""
    hand = Hand('K2.86.QJT7.98642')
    dealer = 'N'
    bids = ['1C', 'P', '4NT', 'P', '5H', 'P', '6S', 'P', 'P', 'D', 'P', 'P', 'P']
    board = _get_board(dealer, hand, bids)
    suit = opening_lead_suit(board)

    assert suit == Suit('C')
    assert suit.selection_reason == '032'



"""Second seat card play."""

import inspect

from bridgeobjects import SUITS, Card, Suit
from bfgdealer import Trick
from ..logger import log

import bfgcardplay.src.global_variables as global_vars
from .player import Player
from .utilities import other_suit_for_signals, get_suit_strength


class SecondSeat():
    def __init__(self, player: Player):
        self.player = player
        self.trick = player.board.tricks[-1]
        self.cards = player.cards_for_trick_suit(self.trick)

    def _select_card_if_void(self, player: Player, trick: Trick) -> Card:
        """Return card if cannot follow suit."""
        trick = self.trick
        manager = global_vars.manager
        # Trump if appropriate
        if (player.trump_suit and
                player.unplayed_cards[player.trump_suit] and
                not player.is_defender):
            if not player.opponents_trumps:
                return log(inspect.stack(), player.unplayed_cards[player.trump_suit][-1])
            elif len(player.total_unplayed_cards[trick.suit.name]) > 7:  # TODO crude
                return log(inspect.stack(), player.unplayed_cards[player.trump_suit][-1])
            else:
                return log(inspect.stack(), player.unplayed_cards[player.trump_suit][0])

        suit = self._best_suit(player)

        # Signal suit preference."""
        if not manager.signal_card[player.seat]:
            suit_cards = player.suit_cards[suit.name]
            for card in suit_cards:
                if not card.is_honour:
                    manager.signal_card[player.seat] = card
                    return log(inspect.stack(), card)

            other_suit = other_suit_for_signals(suit)
            other_suit_cards = player.suit_cards[other_suit]
            if other_suit_cards:
                manager.signal_card[player.seat] = other_suit_cards[-1]
                return log(inspect.stack(), other_suit_cards[-1])

            for suit_name in SUITS:
                if suit_name != suit.name and suit_name != other_suit:
                    final_suit_cards = player.suit_cards[suit_name]
                    if final_suit_cards:
                        return log(inspect.stack(), final_suit_cards[-1])

        for suit_name in SUITS:
            if player.unplayed_cards[suit]:
                card_count = len(player.unplayed_cards[suit])
                if card_count > len(player.total_unplayed_cards[suit]) - card_count:
                    return log(inspect.stack(), player.unplayed_cards[suit][-1])
        return log(inspect.stack(), player.suit_cards[suit.name][-1])

    def _best_suit(self, player: Player) -> Suit:
        """Select suit for signal."""
        # TODO handle no points and equal suits
        cards = player.hand_cards.list
        suit_points = get_suit_strength(cards)
        max_points = 0
        best_suit = None
        for suit in SUITS:
            hcp = suit_points[suit]
            if hcp > max_points:
                max_points = hcp
                best_suit = suit
        if not best_suit:
            return player.longest_suit
        return Suit(best_suit)



from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('first_seat_defender.pbn')


def test_play_partners_suit():
    """Return partner's suit in _select_suit_for_nt_contract."""
    board = boards['0']
    assert next_card(board).name == '8H'


def test_play_winners():
    """Play winners in _select_card_from_suit."""
    board = boards['1']
    assert next_card(board).name == 'JH'


def test_play_winners_not_in_hand():
    """Play winners in _select_card_from_suit."""
    board = boards['2']
    assert next_card(board).name == '4D'


def test_play_winners_in_hand():
    board = boards['3']
    assert next_card(board).name == 'AS'



"""Methods to support Defender play."""
from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import Card, Suit, CARD_VALUES, SUITS

import bfgcardplay.src.global_variables as global_vars
from .utilities import other_suit_for_signals
from .player import Player

MODULE_COLOUR = 'cyan'


def deduce_partner_void_in_trumps(player: Player) -> bool:
    """Return True if we can deduce that partner is void in trumps."""
    suit = player.trump_suit.name
    auction = player.board.auction
    opening_call = auction.seat_calls[player.declarer][0]
    if opening_call.level == 3 and opening_call.denomination == player.trump_suit:
        my_trumps = player.hand.cards_by_suit[suit]
        dummys_trumps = player.hands[player.dummy].cards_by_suit[suit]
        if len(my_trumps) + len(dummys_trumps) >= 5:
            return True
    return False


def dummy_is_short_trump_hand(player: Player) -> bool:
    """Return True if dummy is short trump hand."""
    return len(player.hands[player.dummy].cards_by_suit[player.trump_suit.name]) <= 4


def get_hilo_signal_card(caller) -> Card:
    """Return a signal card denoting even/odd"""
    player = caller.player
    cards = caller.cards
    manager = global_vars.manager
    trick = player.board.tricks[-1]
    if len(cards) % 2 == 0:
        if len(cards) == 2:
            if not cards[-2].is_honour and cards[-2].rank != 'T':
                manager.set_even_odd(player.seat, trick.suit.name, 0)
                return log(inspect.stack(), cards[-2])
            return log(inspect.stack(), cards[-1])

    manager.set_even_odd(player.seat, trick.suit.name, 1)
    return log(inspect.stack(), cards[-1])


def signal_card(player: Player, manager: object, suit: Suit) -> Card | None:
    """Signal suit preference."""
    if not manager.signal_card[player.seat]:
        suit_cards = player.suit_cards[suit.name]
        for card in suit_cards:
            if not card.is_honour:
                manager.signal_card[player.seat] = card
                return log(inspect.stack(), card)

        other_suit = other_suit_for_signals(suit)
        other_suit_cards = player.suit_cards[other_suit]
        if other_suit_cards:
            manager.signal_card[player.seat] = other_suit_cards[-1]
            return log(inspect.stack(), other_suit_cards[-1])

        for suit_name in SUITS:
            if suit_name != suit.name and suit_name != other_suit:
                final_suit_cards = player.suit_cards[suit_name]
                if final_suit_cards:
                    return log(inspect.stack(), final_suit_cards[-1])
    return None


def surplus_card(player: Player) -> Card | None:
    # Discard if more cards than the opposition
    for suit_name in SUITS:
        if player.unplayed_cards[suit_name]:
            card_count = len(player.unplayed_cards[suit_name])
            if card_count > len(player.total_unplayed_cards[suit_name]) - card_count:
                return log(inspect.stack(), player.unplayed_cards[suit_name][-1])
    return None


def best_discard(player: Player) -> Card:
    # Discard from longest_suit, but match suit lengths to dummy
    manager = global_vars.manager
    longest_suit = player.longest_suit
    longest_suit_cards = player.suit_cards[longest_suit]
    if player.trump_suit:
        if len(longest_suit_cards) > len(player.dummys_unplayed_cards[longest_suit]):
            return log(inspect.stack(), longest_suit_cards[-1])

    suits = [suit for suit in SUITS]
    if not player.trump_suit:
        if manager.signal_card[player.seat]:
            suit = manager.signal_card[player.seat].suit.name
            suits.remove(suit)
    for suit in suits:
        cards = player.suit_cards[suit]
        if (len(cards) > len(player.dummys_unplayed_cards[suit]) or
                len(cards) > 1 and cards[0].value == CARD_VALUES['A']):
            return log(inspect.stack(), cards[-1])

    # Last resort - longest_suit
    return log(inspect.stack(), longest_suit_cards[-1])



# first_seat_declarer_nt.py
""" First seat card play for declarer in NT contracts."""

import random
import inspect

from ..logger import log

from bridgeobjects import SUITS, Card, Suit

import bfgcardplay.src.declarer_play as declarer
import bfgcardplay.src.global_variables as global_vars
from .utilities import (get_list_of_best_scores, get_suit_strength, play_to_unblock,
                        get_list_of_best_suits)
from .player import Player, Cards
from .first_seat import FirstSeat
from .declarer_play import card_combinations


class FirstSeatDeclarerNT(FirstSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the card for the first seat."""
        player = self.player
        self.update_manager()
        if len(player.hand.unplayed_cards) == 1:
            return log(inspect.stack(), player.hand.unplayed_cards[0])

        # dashboard = player.dashboard
        # threats = dashboard.threats
        # winners = dashboard.winners
        # suit_lengths = dashboard.suit_lengths
        # extra_winners_length = dashboard.extra_winners_length
        # extra_winners_strength = dashboard.extra_winners_strength

        card = self._select_card_from_manager_suit_to_develop()
        if card:
            return card

        # No strategy established yet so select a suit
        suit = self._select_suit()
        card = self._select_lead_card(suit)
        if card:
            return card

        raise ValueError(f'No cards for {suit}')

    def _select_card_from_manager_suit_to_develop(self) -> Card:
        """Return the selected card from suit to develop."""
        player = self.player
        manager = global_vars.manager
        suit_to_develop = manager.suit_to_develop(player.seat)

        if not suit_to_develop:
            return None

        if player.suit_cards[suit_to_develop]:
            card = card_combinations(player, suit_to_develop.name)
            if card:
                return card

        # Card from suit to develop
        card = self._card_from_suit_to_develop(suit_to_develop)
        if card:
            return card

        return None

    def _card_from_suit_to_develop(self, suit) -> Card:
        """Return card from suit to develop."""
        player = self.player
        manager = global_vars.manager
        if not player.our_unplayed_cards[suit.name]:
            manager.set_suit_to_develop(player.seat, None)
            return None

        if player.suit_cards[suit]:
            card = self._select_card_from_suit_to_develop(suit)
            if card:
                return card
        return None

    def _select_card_from_suit_to_develop(self, suit) -> Card:
        # Play card if suit has missing honours
        card = self._missing_honour(suit)
        if card:
            return card

        # Play a card if suit to develop
        card = self._develop_suit(suit)
        if card:
            return card

        # Play a card if suit has entries
        card = self._card_from_suit_with_entries(suit)
        if card:
            return card

        return None

    def _missing_honour(self, suit) -> Card:
        player = self.player
        manager = global_vars.manager
        if manager.suit_strategy(player.seat)[suit.name] == 'missing_ace':
            card = declarer.lead_to_find_missing_ace(player, suit)
            if card:
                return card
        elif manager.suit_strategy(player.seat)[suit.name] == 'missing_king':
            card = declarer.missing_king(player, suit)
            if card:
                return card
        elif manager.suit_strategy(player.seat)[suit.name] == 'missing_queen':
            card = declarer.missing_queen(player, suit)
            if card:
                return card
        return None

    def _card_from_suit_with_entries(self, suit) -> Card:
        player = self.player
        entries = player.get_entries(player.partners_hand)
        if entries:
            for suit_name, cards in entries.items():
                if cards:
                    if player.suit_cards[suit_name]:
                        card = player.suit_cards[suit_name][-1]
                        return log(inspect.stack(), card)
        return None

    def _develop_suit(self, suit) -> Card:
        player = self.player
        manager = global_vars.manager
        if (len(player.suit_cards[suit]) == 1 and
                not player.partners_suit_cards[suit]):
            card = player.suit_cards[suit][0]
            manager.set_suit_to_develop(player.seat, None)
            return log(inspect.stack(), card)
        return None

    def _select_suit(self) -> Suit:
        """Return the trick lead suit for the declarer in  a suit contract."""
        player = self.player
        manager = global_vars.manager
        if manager.suit_to_develop(player.seat):
            self._check_suit_to_develop()

        # Look for suit to develop
        # cprint(f'{dashboard.tricks_needed=}', MODULE_COLOUR)
        # cprint(f'{dashboard.sure_tricks.count=}', MODULE_COLOUR)
        # cprint(f'{dashboard.sure_tricks.cards=}', MODULE_COLOUR)
        # cprint(f'{player.declarers_tricks=}', MODULE_COLOUR)
        # cprint(f'{dashboard.probable_tricks.cards=}', MODULE_COLOUR)
        # cprint(f'{dashboard.possible_tricks.cards=}', MODULE_COLOUR)
        # # cprint(f'{dashboard.possible_promotions.cards=}', MODULE_COLOUR)

        # Play last winner in case no entry
        suit = self._play_winner_no_entry()
        if suit:
            return suit

        # suit_to_develop has been defined
        suit = manager.suit_to_develop(player.seat)
        if suit and player.unplayed_cards[suit]:
            return log(inspect.stack(), suit)

        # Look for suit to develop
        suit = self._select_suit_to_develop()
        if suit and player.unplayed_cards[suit]:
            return suit

        # This crude logic stops the player cashing their winners early
        suit = self._select_suit_with_winners()
        if suit and player.unplayed_cards[suit]:
            return suit

        suit = self._cross_to_partner()
        if suit:
            return suit

        # Develop longer suits as hands develop
        suit = self._develop_suit_as_play_progresses()
        if suit and player.unplayed_cards[suit]:
            return suit
        # Find suit, but avoid suit if better for opponents to lead
        # suit = self._avoid_suit_with_potential_winner()
        # if suit and player.unplayed_cards[suit]:
        #     return suit

        # Find a suit to develop
        suit = self._find_suit_to_develop()
        if suit and player.unplayed_cards[suit]:
            return suit

        # Find best suit
        return self._select_best_suit()

    def _cross_to_partner(self):
        """Cross to partner if that will force discards."""
        player = self.player
        transfer_hand = False
        for suit in SUITS:
            if (len(player.our_unplayed_cards[suit]) > 0 and
                not player.opponents_unplayed_cards[suit]):
                transfer_hand = True
                if player.unplayed_cards[suit]:
                    return Suit(suit)

        # Transfer to other hand
        if transfer_hand:
            for suit in SUITS:
                if (player.partners_unplayed_cards[suit] and
                        player.partners_unplayed_cards[suit][0] in player.winners[suit] and
                        player.unplayed_cards[suit]):
                    return Suit(suit)
        return None


    def _check_suit_to_develop(self):
        player = self.player
        manager = global_vars.manager
        suit = manager.suit_to_develop(player.seat)
        card = player.unplayed_cards[suit][0]
        if not player.is_master_card(card):
            manager.set_suit_to_develop(player.seat, None)

    def _play_winner_no_entry(self):
        player = self.player
        tricks_needed = player.tricks_needed - player.declarers_tricks
        if tricks_needed <= player.sure_tricks.count:
            for suit in SUITS:
                total_cards = len(player.total_unplayed_cards[suit])
                our_cards = len(player.our_unplayed_cards[suit])
                opponents_length = total_cards - our_cards
                if opponents_length <= 2:
                    if player.unplayed_cards[suit] and not player.partners_unplayed_cards[suit]:
                        if player.is_winner_declarer(player.unplayed_cards[suit][0]):
                            return log(inspect.stack(), Suit(suit))
        return None

    def _select_suit_to_develop(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        extras_suits = self._find_suits_with_extras()

        if extras_suits:
            suits = {suit: 0 for suit in SUITS}
            for suit in extras_suits:
                suits[suit] = len(extras_suits[suit])
            best_suit = get_list_of_best_scores(suits)
            if len(best_suit) == 1:
                suit = Suit(best_suit[0])
                manager.set_suit_to_develop(player.seat, suit)
                if player.unplayed_cards[suit]:
                    return log(inspect.stack(), suit)
        return None

    def _select_best_suit(self) -> Suit:
        player = self.player
        suit_scores = {suit: 0 for suit in SUITS}
        if len(player.board.tricks) >= 5:
            for suit_name, suit in SUITS.items():
                if suit != player.trump_suit:
                    if player.holds_all_winners_in_suit(suit):
                        suit_scores[suit_name] = 1
        total = 0
        for score in suit_scores.values():
            total += score
        if not total:
            for suit in SUITS:
                our_cards = Cards(player.our_unplayed_cards[suit])
                suit_scores[suit] += our_cards.value + len(player.our_unplayed_cards[suit])
        suits = get_list_of_best_scores(suit_scores)
        if len(suits) == 1 and len(player.unplayed_cards[suits[0]]) > 0:
            return log(inspect.stack(), Suit(suits[0]))
        suit_choices = [suit for suit in SUITS if len(player.unplayed_cards[suit]) > 0]
        suit = Suit(random.choice(suit_choices))
        return log(inspect.stack(), suit)

    def _find_suit_to_develop(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        if not manager.suit_to_develop(player.seat):
            suit_to_develop = self._suit_to_develop()
            if suit_to_develop:
                return log(inspect.stack(), suit_to_develop)
        return None

    def _find_suits_with_extras(self) -> list[Suit]:
        player = self.player
        if player.probable_tricks.count:
            return player.probable_tricks.suits
        return player.possible_tricks.suits

    def _select_suit_with_winners(self):
        """Return"""
        player = self.player
        manager = global_vars.manager

        set_strategy_limit = 1
        if len(player.board.tricks) < 4:
            set_strategy_limit = 7

        # Play
        best_suits = get_list_of_best_suits(player)
        winners = {suit: len(player.winners[suit]) for suit in SUITS}
        # best_suits = get_list_of_best_scores(winners)
        for suit in best_suits:
            if (player.winners[suit] and
                    player.unplayed_cards[suit] and
                    len(player.our_cards[suit]) >= set_strategy_limit):
                manager.set_suit_strategy(player.seat, suit, 'Play winners')
                return log(inspect.stack(), Suit(suit))
        return None

    def _develop_suit_as_play_progresses(self):
        """Return"""
        player = self.player
        manager = global_vars.manager
        candidate_suits = {}
        if len(player.board.tricks) > 1:
            for suit in SUITS:
                if (len(player.unplayed_cards[suit]) >= 4 or
                        (len(player.partners_unplayed_cards[suit]) >= 4 and
                         len(player.unplayed_cards[suit]) > 0)):
                    points = get_suit_strength(player.our_unplayed_cards[suit])
                    candidate_suits[suit] = len(player.our_unplayed_cards[suit]) + points[suit]
            if candidate_suits:
                candidates = get_list_of_best_scores(candidate_suits)
                suit = candidates[0]
                manager.set_suit_to_develop(player.seat, Suit(suit))
                for card in player.partners_unplayed_cards[suit]:
                    if card.is_honour and player.is_winner_declarer(card):
                        return log(inspect.stack(), Suit(suit))
        return None

    def _avoid_suit_with_potential_winner(self):
        """Return"""
        player = self.player
        suits = []
        for suit in SUITS:
            if player.unplayed_cards[suit]:
                if not self._potential_winner(suit):
                    suits.append(suit)

        # Must have one suit!
        if not suits:
            for suit in SUITS:
                if player.unplayed_cards[suit]:
                    suits.append(suit)

        if len(suits) == 1:
            return log(inspect.stack(), Suit(suits[0]))
        return None

    def _suit_to_develop(self) -> list[str]:
        """Return a list of suits to develop."""
        player = self.player
        if len(player.board.tricks) > 9:
            return None
        manager = global_vars.manager
        long_suits_to_develop = self._long_suits_to_develop()
        longest_strongest_suit = self._longest_strongest_suit(long_suits_to_develop)
        suit = Suit(longest_strongest_suit)
        manager.set_suit_to_develop(player.seat, suit)
        suit_strategy = self._set_suit_strategy(suit)
        manager.set_suit_strategy(player.seat, suit.name, suit_strategy)
        return suit

    def _set_suit_strategy(self, suit: Suit) -> str:
        """Return the suit strategy as a string."""
        player = self.player

        missing_honours = player.missing_honours(suit)
        for missing_honour in missing_honours:
            if missing_honour.rank == 'A':
                return 'missing_ace'
            if missing_honour.rank == 'K':
                return 'missing_king'
            if missing_honour.rank == 'Q':
                return 'missing_queen'
        return ''

    def _long_suits_to_develop(self):
        """Return a list of long suits to develop."""
        player = self.player
        long_suits_to_develop = []
        for suit_name in SUITS:
            declarers_suit_cards = player.declarers_suit_cards[suit_name]
            dummys_suit_cards = player.dummys_suit_cards[suit_name]
            total_cards = len(declarers_suit_cards) + len(dummys_suit_cards)
            if (total_cards >= 6 and (len(declarers_suit_cards) >= 4
                                      or len(dummys_suit_cards) >= 4)):
                long_suits_to_develop.append(suit_name)
        return long_suits_to_develop

    def _longest_strongest_suit(self, suits: list[str]) -> str:
        """Return the longest and strongest suit from a list of suit_names."""
        player = self.player
        manager = global_vars.manager
        candidate_suits = {}
        for suit_name in suits:
            declarers_suit_cards = player.declarers_suit_cards[suit_name]
            dummys_suit_cards = player.dummys_suit_cards[suit_name]
            candidate_suits[suit_name] = len(declarers_suit_cards) + len(dummys_suit_cards)
        long_suits = get_list_of_best_scores(candidate_suits)

        if len(long_suits) == 1:
            suit = long_suits[0]
            manager.set_suit_to_develop(player.seat, suit)
            suit_strategy = self._set_suit_strategy(Suit(suit))
            manager.set_suit_strategy(player.seat, suit, suit_strategy)
            return suit

        strong_suits = get_list_of_best_scores(candidate_suits)
        if len(strong_suits) == 1:
            return log(inspect.stack(), strong_suits[0])

        if strong_suits:
            suit = random.choice(strong_suits)
            return log(inspect.stack(), Suit(suit))
        for suit in SUITS:
            if player.winners[suit] and player.unplayed_cards[suit]:
                return log(inspect.stack(), suit)

        suits = []
        for suit in SUITS:
            if player.unplayed_cards[suit]:
                suits.append(suit)
        suit = random.choice(suits)
        return suit

    def _select_lead_card(self, suit: Suit) -> Card | None:
        """Return the selected lead card for declarer."""
        player = self.player
        manager = global_vars.manager
        cards = player.suit_cards[suit.name]
        if not cards:
            return None

        # Card combinations
        card = card_combinations(player, suit.name)
        if card:
            return log(inspect.stack(), card)

        # Play winning cards
        if (manager.suit_strategy(player.seat)[suit.name] == 'Play winners' or
                player.holds_all_winners_in_suit(suit)):
            my_cards = player.unplayed_cards[suit.name]
            partners_cards = player.partners_unplayed_cards[suit.name]
            card = play_to_unblock(my_cards, partners_cards)
            if card:
                return log(inspect.stack(), card)

            if (partners_cards and len(my_cards) > len(partners_cards) and
                    player.is_winner_declarer(partners_cards[0]) and
                    len(partners_cards) > 1):
                return log(inspect.stack(), my_cards[-1])
            for card in cards[::-1]:
                if player.opponents_unplayed_cards[suit.name]:
                    if player.is_winner_declarer(card):
                        return log(inspect.stack(), card)
                else:
                    return log(inspect.stack(), card)

        # Top of touching honours
        for index, card in enumerate(cards[:-1]):
            if card.is_honour and card.value == cards[index+1].value + 1:
                return log(inspect.stack(), card)

        # Top of doubleton
        if len(cards) == 2:
            return log(inspect.stack(), cards[0])

        # Return bottom card
        return log(inspect.stack(), cards[-1])

    def _potential_winner(self, suit: str) -> bool:
        """Return True if the suit has a potential winner."""
        player = self.player
        for cards in [player.unplayed_cards[suit], player.partners_unplayed_cards[suit]]:
            if (cards and
                    cards[0] != player.total_unplayed_cards[suit][0] and
                    cards[0] == player.total_unplayed_cards[suit][1] and
                    len(cards) >= 2):
                return True
        return False



""" Opening leads for Card Player."""

from .opening_lead_card import opening_lead_card
from .opening_lead_suit import opening_lead_suit

import bfgcardplay.src.global_variables as global_vars

MODULE_COLOUR = 'green'


def opening_lead(board):
    """Return the proposed opening lead for the board."""

    manager = global_vars.manager
    manager.initialise()

    if not board.contract.declarer:
        return None
    opening_suit = opening_lead_suit(board)
    cards = [card for card in board.hands[board.contract.leader].cards if card.suit == opening_suit]
    opening_card = opening_lead_card(cards, board.contract)
    return (opening_suit, opening_card)



from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('second_seat_defender.pbn')


def test_play_partners_suit():
    """Do not play high card when inappropriate."""
    board = boards['0']
    assert next_card(board).name != 'KH'


def test_play_discard():
    board = boards['1']
    assert next_card(board).name != 'KC'
    assert next_card(board).name != '9D'



"""Second seat card play for declarer."""

from bridgeobjects.src.constants import CARD_VALUES

from termcolor import cprint

import inspect

from ..logger import log

from bridgeobjects import Card, SUITS, Suit
from .player import Player
from .utilities import get_list_of_best_scores, get_suit_strength
from .second_seat import SecondSeat
import bfgcardplay.src.global_variables as global_vars

MODULE_COLOUR = 'green'


class SecondSeatDeclarer(SecondSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the card if the second seat."""
        player = self.player
        trick = self.trick
        cards = self.cards

        # Void
        if not cards:
            return self._select_card_if_void()

        # Singleton
        if len(cards) == 1:
            return log(inspect.stack(), cards[0])

        # Win trick if partner cannot
        card = self._win_trick_if_partner_cannot()
        if card:
            return log(inspect.stack(), card)

        # Look for tenace about a threat card
        (higher_card, lower_card) = player.card_from_tenace_threat()
        if lower_card:
            return log(inspect.stack(), lower_card)

        # Does a tenace exist
        lower_card = player.card_from_tenace()
        if lower_card:
            return log(inspect.stack(), lower_card)

        # Card from suit to develop
        card = self._card_from_suit_to_develop()
        if card:
            return log(inspect.stack(), card)

        # When to duck - rule of 7
        card = self._do_not_duck_if_appropriate()
        if card:
            return log(inspect.stack(), card)

        # Cover honour if own touching honours
        if trick.cards[0].is_honour:
            card = self._cover_honour()
            if card:
                return log(inspect.stack(), card)

        # Play card if it can beat intermediate lead
        card = self._win_over_intermediate_lead()
        if card:
            return log(inspect.stack(), card)

        # Partner is void or only has pips
        card = self._partner_weak()
        if card:
            return log(inspect.stack(), card)

        # Play lowest card
        return log(inspect.stack(), cards[-1])

    def _win_trick_if_partner_cannot(self) -> Card:
        if not self.player.board.contract.denomination.is_suit:
            return None
        player = self.player
        trick = self.trick
        partners_cards = player.partners_unplayed_cards[trick.suit.name]
        if partners_cards and not partners_cards[0].is_honour:
            for card in self.cards[len(partners_cards)-1::-1]:
                if player.is_winner_declarer(card, trick):
                    return log(inspect.stack(), card)
        return None

    def _do_not_duck_if_appropriate(self) -> Card:
        cards = self.cards
        duck_trick = self._can_duck_trick()
        if not duck_trick:
            if cards[0].value > self.player.board.tricks[-1].cards[0].value:
                return log(inspect.stack(), cards[0])
        return None

    def _can_duck_trick(self) -> bool:
        player = self.player
        if not player.trump_suit:
            return self._duck_trick()
        return False

    def _duck_trick(self) -> bool:
        """Return True if the player is to duck the trick."""
        player = self.player
        trick = self.trick
        cards = self.cards
        opponents_unplayed_cards = player.opponents_unplayed_cards[trick.suit.name]
        if cards and opponents_unplayed_cards:
            partners_cards = player.partners_suit_cards[trick.suit.name]
            partner_can_win = False
            if partners_cards:
                if partners_cards[0].value > trick.cards[0].value:
                    partner_can_win = True
            can_win_trick = (cards[0].value > opponents_unplayed_cards[0].value and
                             cards[0].value > trick.cards[0].value and
                             not partner_can_win)
            if self._rule_of_seven() and can_win_trick:
                return False
        return True

    def _cover_honour(self) -> Card:
        player = self.player
        trick = self.trick
        cards = self.cards

        # Cover honour if own touching honours
        touching_honours = player.touching_honours(player.our_unplayed_cards[trick.suit.name])
        my_unplayed_cards = len(player.unplayed_cards[trick.suit.name])
        partners_unplayed_cards = len(player.partners_unplayed_cards[trick.suit.name])
        if (trick.cards[0].value >= CARD_VALUES['T'] and
                trick.cards[0].value < cards[0].value and
                cards[0].is_honour and
                touching_honours and
                my_unplayed_cards <= partners_unplayed_cards):
            return log(inspect.stack(), cards[0])

        # TODO see this web site http://www.rpbridge.net/4l00.htm
        duck_trick = self._can_duck_trick()
        if trick.cards[0].value >= CARD_VALUES['T'] and not duck_trick:
            for card in player.suit_cards[trick.suit.name][::-1]:
                if card.value > trick.cards[0].value:
                    return log(inspect.stack(), card)

        if (len(cards) > 1 and
                cards[0].value == cards[1].value + 1 and
                CARD_VALUES['A'] > cards[1].value >= CARD_VALUES['T']):
            return log(inspect.stack(), cards[1])
        return None

    def _win_over_intermediate_lead(self) -> Card:
        trick = self.trick
        cards = self.cards
        if trick.cards[0].value >= CARD_VALUES['9']:  # nine or above
            if len(cards) >= 2:
                if cards[1].value >= CARD_VALUES['T']:
                    for card in cards[::-1]:
                        if card.value > trick.cards[0].value:
                            return log(inspect.stack(), card)
        return None

    def _partner_weak(self) -> Card:
        player = self.player
        if not player.board.contract.denomination.is_suit:
            return None
        trick = self.trick
        partner_void = True
        if player.partners_unplayed_cards[trick.suit.name]:
            partner_void = False
            partners_top_card = player.partners_unplayed_cards[trick.suit.name][0]
            partner_honour = partners_top_card.is_honour

        if partner_void or not partner_honour:
            for card in self.cards[::-1]:
                if card.is_honour and card.value > trick.cards[0].value:
                    return log(inspect.stack(), card)
        return None

    def _rule_of_seven(self) -> bool:
        """Return True if rule of seven applies."""
        trick = self.trick
        player = self.player
        our_cards = player.our_cards[trick.suit]
        duck_count = 7 - len(our_cards) - len(player.board.tricks)
        return duck_count < 0

    def _card_from_suit_to_develop(self):
        """Return card from a suit to develop."""
        player = self.player
        manager = global_vars.manager
        trick = player.board.tricks[-1]
        suit_to_develop = manager.suit_to_develop(player.seat)
        if suit_to_develop:
            for card in player.unplayed_cards[trick.suit.name][::-1]:
                if player.is_winner_declarer(card, trick):
                    return log(inspect.stack(), card)
        return None

    def _select_card_if_void(self) -> Card:
        """Return card if cannot follow suit."""
        player = self.player
        trick = player.board.tricks[-1]

        card = self._trump_if_appropriate()
        if card:
            return card

        for suit_name in SUITS:
            if player.unplayed_cards[suit_name]:
                card_count = len(player.unplayed_cards[suit_name])
                if card_count > len(player.total_unplayed_cards[suit_name]) - card_count:
                    return log(inspect.stack(), player.unplayed_cards[suit_name][-1])

        lengths = {suit: 0 for suit in SUITS}
        for suit in SUITS:
            if player.trump_suit and suit == player.trump_suit.name:
                lengths[suit] = 0
            else:
                lengths[suit] = len(player.unplayed_cards[suit])

        long_suits = get_list_of_best_scores(lengths)
        if player.unplayed_cards[long_suits[0]]:
            return log(inspect.stack(), player.unplayed_cards[long_suits[0]][-1])

        return log(inspect.stack(), player.unplayed_cards[player.trump_suit.name][-1])

    def _trump_if_appropriate(self) -> Suit | None:
        """Trump if appropriate."""
        player = self.player
        trick = player.board.tricks[-1]
        suit_name = trick.suit.name

        if player.trump_suit and player.unplayed_cards[player.trump_suit]:
            defender_is_void = (player.voids[player.right_hand_seat][suit_name] or
                                player.voids[player.left_hand_seat][suit_name])
            if len(player.opponents_unplayed_cards[suit_name]) >= 4 and not defender_is_void:
                trump_low = True
            else:
                trump_low = False

            partners_cards = player.partners_unplayed_cards[suit_name]
            partners_trumps = player.partners_unplayed_cards[player.trump_suit.name]
            my_trumps = player.unplayed_cards[player.trump_suit]
            if not partners_cards and partners_trumps:
                if len(my_trumps) > len(partners_trumps):
                    return None

            if not partners_cards or not player.is_winner_declarer(partners_cards[0], trick):
                if not player.opponents_trumps:
                    return log(inspect.stack(), my_trumps[-1])
                elif trump_low:
                    return log(inspect.stack(), my_trumps[-1])
                else:
                    return log(inspect.stack(), my_trumps[0])
        return None

    def _best_suit(self, player: Player) -> Suit:
        """Select suit for signal."""
        # TODO handle no points and equal suits
        cards = player.hand_cards.list
        suit_points = get_suit_strength(cards)
        max_points = 0
        best_suit = None
        for suit in SUITS:
            if player.trump_suit and suit != player.trump_suit.name:
                hcp = 0
            else:
                hcp = suit_points[suit]
                if hcp > max_points:
                    max_points = hcp
                    best_suit = suit
        if best_suit:
            return Suit(best_suit)
        return player.longest_suit




import random

from bridgeobjects import Suit, Card, RANKS, SUITS
from bfgbidding import Hand

from .utilities import get_list_of_best_scores


MODULE_COLOUR = 'yellow'


class Cards():
    """A class representing a set of cards."""
    def __init__(self, input: list[str] | list[Card] | Hand):
        self.list = []
        self.longest_suit = None
        by_suit = {suit: [] for key, suit in SUITS.items()}
        by_suit_name = {key: [] for key, suit in SUITS.items()}
        self.by_suit = {**by_suit, **by_suit_name}
        self._value = 0
        self._create(input)

    def count(self) -> int:
        """Return the number of cards."""
        return len(self.list)

    def __repr__(self):
        return f'Cards: {self.by_suit}'

    def _create(self, input: list[str] | list[Card] | Hand):
        """Create a set of cards"""
        if isinstance(input, Hand):
            cards = input.cards
        elif not isinstance(input, list):
            raise TypeError('Cards must be a list of cards')
        else:
            cards = input

        raw_list = []
        for card in cards:
            if isinstance(card, str):
                card = Card(card)
            raw_list.append(card)
        self.list = self.sort_cards(raw_list)

        self._value = 0
        for card in self.list:
            suit = card.suit
            self.by_suit[suit].append(card)
            self.by_suit[suit.name].append(card)
            self._value += card.value
        self.longest_suit = self._longest_suit()

    @staticmethod
    def sort_cards(cards: list[Card]) -> list[Card]:
        """Return a sorted list of cards."""
        return sorted(cards, reverse=True)

    def _longest_suit(self) -> Suit:
        """Return the suit with most cards."""
        suit_dict = {}
        for suit, card_list in self.by_suit.items():
            if isinstance(suit, Suit):
                suit_dict[suit] = len(card_list)
        suits = get_list_of_best_scores(suit_dict)
        suit = random.choice(suits)
        return suit

    @property
    def value(self) -> int:
        """Return the sum of the values of the cards."""
        return self._value


class SuitCards():
    """Represent all of the cards in a suit by name."""
    def __init__(self, suit: str):
        self.suit = suit
        self.ace = Card('A', suit)
        self.king = Card('K', suit)
        self.queen = Card('Q', suit)
        self.jack = Card('J', suit)
        self.ten = Card('T', suit)
        self.nine = Card('9', suit)
        self.eight = Card('8', suit)
        self.seven = Card('7', suit)
        self.six = Card('6', suit)
        self.five = Card('5', suit)
        self.four = Card('4', suit)
        self.three = Card('3', suit)
        self.two = Card('2', suit)

    def cards(self, top_rank: str = '', bottom_rank: str = '') -> tuple[Card]:
        """Return a tuple of the cards requested."""
        if not top_rank:
            top_rank = 'A'
        if not bottom_rank:
            bottom_rank = '2'

        cards = []
        ranks = [rank for rank in RANKS]
        ranks.reverse()
        ranks = ranks[:-1]
        top_index = ranks.index(top_rank)
        bottom_index = ranks.index(bottom_rank)
        requested_ranks = ranks[top_index:bottom_index+1]
        for rank in requested_ranks:
            cards.append(Card(rank, self.suit))
        return tuple(cards)

    def sorted_cards(self):
        """Return a list of Cards in descending order by rank."""
        ranks = [rank for rank in RANKS[1:]]
        ranks.reverse()
        cards = [Card(rank, self.suit) for rank in ranks]
        return cards


class CardArray():
    """Data structure for dashboard analyses"""
    def __init__(self, cards: list[Card]):
        self.cards = self.sort_cards(cards)
        self.suits = self._suits_from_cards()

    def _suits_from_cards(self):
        """Return a dict of suit cards."""
        suits = {suit: [] for suit in SUITS}
        for card in self.cards:
            suits[card.suit.name].append(card)
        return suits

    @property
    def count(self):
        """Return the length of cards"""
        return len(self.cards)

    def __repr__(self):
        # cprint(f'{self.cards}', MODULE_COLOUR)
        # cprint(f'{self.suits}', MODULE_COLOUR)
        return ''

    @staticmethod
    def sort_cards(cards: list[Card]) -> list[Card]:
        """Return a sorted list of cards."""
        return sorted(cards, reverse=True)



from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('first_seat_declarer.pbn')


def test_select_suit_with_cards():
    board = boards['0']
    next_card(board).name  # Seem to need to play this card before we get the correct card
    assert next_card(board).name[1] != 'D'


def test_select_suit_with_winners_in_partners_hand():
    board = boards['1']
    assert next_card(board).name[1] == 'S'


def test_do_not_play_losers_when_you_have_winners():
    board = boards['2']
    assert next_card(board).suit.name != 'C'


def test_play_winners():
    # see first_seat_declarer_nt.py # Play winning cards at 443
    board = boards['3']
    # TODO this causes a problem in test second_seat_declarer
    # assert next_card(board).name != '7D'


def test_underplay_tenace():
    board = boards['4']
    print(next_card(board))
    assert next_card(board).name[1] != ''



"""Provide the Dashboard class for bfgcardplay."""

from termcolor import cprint
from dataclasses import dataclass

from bridgeobjects import Hand, Card, Suit, SUITS, RANKS
from .data_classes import SuitCards, CardArray

SPLIT_PROBABILITIES = {
    '3-3': 35.5,
    '4-2': 48.3,
    '5-1': 14.5,
    '6-0': 1.5,
    '3-2': 67.8,
    '4-1': 28.3,
    '5-0': 3.9,
    '3-1': 49.7,
    '2-2': 40.7,
    '4-0': 9.5,
    '2-1': 78,
    '3-0': 22,
    '1-1': 52,
    '2-0': 48,
}

SPLITS = {
    6: ['3-3', '4-2', '5-1', '6-0'],
    5: ['3-2', '4-1', '5-0'],
    4: ['3-1', '2-2', '4-0'],
    3: ['2-1', '3-0'],
    2: ['1-1', '2-0'],
}


COLOUR = 'blue'


@dataclass
class SplitWinners:
    split: str
    probability: float
    cards: str


class Dashboard():
    "Statistics for a player."
    def __init__(self, player) -> None:
        self.suits = {suit: {} for suit in SUITS}
        self.player = player
        # self.threats = self._get_threats()
        # self.losers = self._get_losers()
        self.winners = self.player.winners
        self.losers = self.player.losers
        self.split_winners = self._get_split_winners()
        self.winner_count = self._get_winner_count()
        self.suit_lengths = self._suit_lengths()
        self.extra_winners_length = self._extra_winners_length()
        # extra_winners_strength must be generated after winners
        self.extra_winners_strength = self._extra_winners_strength()
        self.tricks_needed = 6 + int(player.board.contract.name[0])

        self.sure_tricks = CardArray(self._get_current_sure_tricks())
        self.probable_tricks = CardArray(self._get_current_probable_tricks())
        self.possible_tricks = CardArray(self._get_current_possible_tricks())
        self.possible_promotions = CardArray(self._get_possible_promotions())

        (my_entries, partners_entries) = self._get_current_entries()
        self.entries = my_entries
        self.partners_entries = partners_entries

    def __repr__(self):
        cprint(f'{"threats":<10}{self.threats}', COLOUR)
        cprint(f'{"winners":<10}{self.winners}', COLOUR)
        cprint(f'{"suit_lens":<10}{self.suit_lengths}', COLOUR)
        # cprint(f'{"threats":<10}{self.extra_winners_length}', COLOUR)
        # cprint(f'{"threats":<10}{self.extra_winners_strength}', COLOUR)
        return ''

    def _get_current_sure_tricks(self) -> list[Card]:
        """Return a list of certain winners based on unplayed cards."""
        player = self.player
        sure_tricks = []
        declarers_cards = player.unplayed_cards
        dummys_cards = player.partners_unplayed_cards
        for suit in SUITS:
            declarer_suit_cards = [card for card in declarers_cards[suit]]
            dummy_suit_cards = [card for card in dummys_cards[suit]]
            suit_cards = declarer_suit_cards + dummy_suit_cards
            our_cards = CardArray.sort_cards(suit_cards)
            our_length = max(len(declarer_suit_cards), len(dummy_suit_cards))
            for card in our_cards[:our_length]:
                if player.is_winner_declarer(card, player.board.tricks[-1]):
                    sure_tricks.append(card)
        return sure_tricks

    def _get_current_probable_tricks(self) -> tuple[dict[str, list[Card]]]:
        """Return a dict of probable winners by
        suit based on unplayed cards."""
        opponents_length = {
            8: 5,
            7: 4,
            6: 4,
            5: 3,
            4: 3,
            3: 2,
            2: 2,
            1: 1,
            0: 0,
        }
        return self._current_trick_calculation(opponents_length)

    def _get_current_possible_tricks(self) -> tuple[dict[str, list[Card]]]:
        """Return a dict of possible winners by suit based
        on unplayed cards."""
        opponents_length = {
            8: 4,
            7: 4,
            6: 3,
            5: 3,
            4: 2,
            3: 2,
            2: 1,
            1: 1,
            0: 0,
        }
        return self._current_trick_calculation(opponents_length)

    def _current_trick_calculation(
            self,
            opponents_length: dict[int, int]
            ) -> tuple[dict[str, list[Card]]]:
        """Return a dict of winners by suit based on unplayed cards."""
        player = self.player
        probable_tricks = []
        declarers_cards = player.unplayed_cards
        dummys_cards = player.partners_unplayed_cards
        opponents_cards = player.opponents_unplayed_cards
        for suit in SUITS:
            declarer_suit_cards = [card for card in declarers_cards[suit]]
            dummy_suit_cards = [card for card in dummys_cards[suit]]
            opponents_holding = (len(player.total_unplayed_cards[suit])
                                 - len(declarer_suit_cards)
                                 - len(dummy_suit_cards))

            if opponents_holding in opponents_length:
                # Honours
                honours = SuitCards(suit).cards('A', 'T')
                opponents_honours = 0
                for index, card in enumerate(honours):
                    if index < opponents_length[opponents_holding]:
                        if card in opponents_cards[suit]:
                            opponents_honours += 1
                opponents_count = min(
                    opponents_length[opponents_holding],
                    opponents_honours)
                opponents_count = opponents_length[opponents_holding]

                # Length
                our_length = max(
                    len(declarer_suit_cards),
                    len(dummy_suit_cards)
                    )
                if len(declarer_suit_cards) > len(dummy_suit_cards):
                    our_long_hand = declarer_suit_cards
                else:
                    our_long_hand = dummy_suit_cards

                if our_long_hand:
                    range_limit = our_length - opponents_count
                    for index in range(range_limit):
                        card = our_long_hand[-index]
                        if card not in self.sure_tricks.cards:
                            probable_tricks.append(card)
        return probable_tricks

    def _get_possible_promotions(self):
        """Return a dict of possible promotions
        by suit based on unplayed cards."""
        def check_card(self, hand, card, buffer):
            """Return True if the card is a candidate for promotion."""
            if (card in hand and len(hand) >= buffer and
                    card not in self.sure_tricks.cards):
                return True
            return False

        player = self.player
        possible_promotions = []
        declarers_cards = player.unplayed_cards
        dummys_cards = player.partners_unplayed_cards
        for suit in SUITS:
            if suit not in self.probable_tricks.suits:
                honours = SuitCards(suit).cards('K', 'T')
                for hand in [declarers_cards[suit], dummys_cards[suit]]:
                    for index, card in enumerate(honours):
                        if check_card(self, hand, card, 2+index):
                            possible_promotions.append(card)
        return possible_promotions

    def _get_current_entries(self) -> tuple[dict[str, list[Card]]]:
        """Return a dict of probable entries by suit based
        on unplayed cards."""
        player = self.player
        my_entries = {suit: [] for suit in SUITS}
        partners_entries = {suit: [] for suit in SUITS}
        for suit in SUITS:
            my_cards = player. unplayed_cards[suit]
            partners_cards = player.partners_unplayed_cards[suit]
            if my_cards and partners_cards:
                if (player.is_winner_declarer(my_cards[0]) and
                        my_cards[0].value > partners_cards[-1].value):
                    my_entries[suit] = my_cards[0]
                if (player.is_winner_declarer(partners_cards[0]) and
                        partners_cards[0].value > my_cards[-1].value):
                    partners_entries[suit] = partners_cards[0]
        return (my_entries, partners_entries)

    # def _get_entries_for_hand(self, my_cards, partners_cards):
    #     player = self.player
    #     entries = {suit: [] for suit in SUITS}
    #     for suit in SUITS:
    #         cards = my_cards[suit]
    #         if cards:
    #             if (player.is_winner_declarer(cards[0]) and
    #                     cards[0].value > partners_cards[suit][-1].value):
    #                 entries[suit] = cards[0]
    #     return entries

    # def _get_threats(self) -> dict[str, Card]:
    #     """Return a dict of cards by suit that threaten players cards."""
    #     if self.player.trump_suit:
    #         return self._get_threats_suit_contract()
    #     else:
    #         return self._get_threats_nt_contract()

    def _get_threats_nt_contract(self) -> dict[str, Card]:
        """Return a dict of cards by suit that threaten players cards."""
        player = self.player

        threats = {suit: [] for suit in SUITS}
        ranks = [rank for rank in RANKS]
        ranks.reverse()
        sorted_ranks = ranks[:-1]

        for suit in SUITS:
            long_cards = player.unplayed_cards[suit]
            our_cards = player.our_unplayed_cards[suit]
            opponents_cards = player.opponents_unplayed_cards[suit]
            missing = len(opponents_cards)
            winners = 0
            skip_card = False
            for index, rank in enumerate(sorted_ranks):
                ranking_card = Card(rank, suit)
                if ranking_card in player.total_unplayed_cards[suit]:
                    if ranking_card in our_cards:
                        if not skip_card:
                            winners += 1
                        skip_card = False

                        # No more cards to lose
                        if winners > missing:
                            break
                    else:
                        threats[suit].append(ranking_card)
                        # skip card means that the next card in the suit is
                        # assumed to have been beaten
                        skip_card = True

                    # threats only in long hand
                    if index + 1 >= len(long_cards):
                        break
                if len(opponents_cards) <= index + 1:
                    break
        return threats

    def _get_threats_suit_contract(self) -> dict[str, Card]:
        """Return a dict of cards by suit that threaten players cards."""
        player = self.player
        (long_hand, short_hand) = self.get_long_short_trump_hands()
        unplayed_cards = player.cards_by_suit(long_hand)
        long_cards = unplayed_cards[player.trump_suit.name]

        threats = {suit: [] for suit in SUITS}
        ranks = [rank for rank in RANKS]
        ranks.reverse()
        sorted_ranks = ranks[:-1]

        for suit in SUITS:
            our_cards = player.our_unplayed_cards[suit]
            my_cards = player.unplayed_cards[suit]
            partners_cards = player.partners_unplayed_cards[suit]
            max_length = max(len(my_cards), len(partners_cards))
            opponents_cards = player.opponents_unplayed_cards[suit]
            missing = len(opponents_cards)
            winners = 0
            skip_card = False
            for index, rank in enumerate(sorted_ranks):
                ranking_card = Card(rank, suit)
                if ranking_card in player.total_unplayed_cards[suit]:
                    if ranking_card in our_cards:
                        if not skip_card:
                            winners += 1
                        skip_card = False

                        # Only count to the number of cards in our hands
                        if index + 1 >= max_length:
                            break

                        # No more cards to lose
                        if winners > missing:
                            break
                    else:
                        threats[suit].append(ranking_card)
                        # skip card means that the next card in the suit is
                        # assumed to have been beaten
                        skip_card = True

                    # threats only in long hand
                    if index + 1 >= len(long_cards):
                        break
                if len(opponents_cards) <= index + 1:
                    break
        return threats

    def get_long_short_trump_hands(self) -> tuple[Hand, Hand]:
        """Return the  hands for long and short trump hand between
        player and partner."""
        return self.get_long_short_hands(self.player.trump_suit)

    def get_long_short_hands(self, suit: Suit) -> tuple[Hand, Hand]:
        """Return the  hands for long and short hand between
        player and partner."""
        player = self.player
        my_unplayed_cards = player.unplayed_cards[suit.name]
        partners_unplayed_cards = player.partners_unplayed_cards[suit.name]
        if len(my_unplayed_cards) >= len(partners_unplayed_cards):
            return (player.hand, player.partners_hand)
        else:
            return (player.partners_hand, player.hand)

    def _get_split_winners(self) -> dict[str, Card]:
        """Return a dict of cards by suit that are
        certain winners depending on splits."""
        player = self.player
        opponents_cards = player.opponents_unplayed_cards
        split_winners = {suit: [] for suit in SUITS}
        for suit in SUITS:
            if len(opponents_cards[suit]) in SPLITS:
                splits = SPLITS[len(opponents_cards[suit])]
                for split in splits:
                    winners = self._split_winners(suit, split)
                    split_winners[suit].append(winners)
        return split_winners

    def _split_winners(self, suit, split):
        player = self.player
        our_cards = player.our_unplayed_cards
        our_length = self._get_our_length(player, suit)
        opponents_cards = player.opponents_unplayed_cards
        split_winners = SplitWinners(split, SPLIT_PROBABILITIES[split], '')
        opps_suit_cards = opponents_cards[suit][:int(split[0])]
        for index, card in enumerate(our_cards[suit]):
            split_winners.cards += self._split_winner(index,
                                                      card, opps_suit_cards,
                                                      split_winners.cards,
                                                      our_length)
        return split_winners

    @staticmethod
    def _split_winner(index, card, opps_suit_cards, cards, our_length):
        if index < len(opps_suit_cards):
            if len(cards) < our_length:
                if opps_suit_cards:
                    if card.value > opps_suit_cards[0].value:
                        return card.rank
        elif index < our_length:
            return card.rank
        return ''

    @staticmethod
    def _get_our_length(player: object, suit: str) -> int:
        my_cards = player.unplayed_cards
        partners_cards = player.partners_unplayed_cards
        if len(my_cards[suit]) > len(partners_cards[suit]):
            return len(my_cards[suit])
        return len(partners_cards[suit])

    def _suit_lengths(self) -> dict[str, int]:
        """Return a dict of suit_lengths."""
        player = self.player
        suit_lengths = {suit: [] for suit in SUITS}
        our_cards = player.our_unplayed_cards
        for suit in SUITS:
            suit_lengths[suit] = len(our_cards[suit])
        return suit_lengths

    def _extra_winners_length(self):
        """Return a dict of potential extra length winners by suit."""
        player = self.player
        splits = {
            6: 4,
            5: 3,
            4: 3,
            3: 2,
            2: 2,
            1: 1,
            0: 0,
        }
        extra_winners = {suit: 0 for suit in SUITS}
        our_cards = player.our_unplayed_cards
        my_cards = player.unplayed_cards
        partners_cards = player.partners_unplayed_cards
        opponents_cards = player.opponents_unplayed_cards
        for suit in SUITS:
            if not len(our_cards[suit]) > len(opponents_cards[suit]):
                break
            our_length = max(len(my_cards[suit]), len(partners_cards[suit]))
            split = splits[len(opponents_cards[suit])]
            extra_winners[suit] = our_length - split
        return extra_winners

    def _extra_winners_strength(self):
        """Return a dict of potential extra strength winners by suit."""
        player = self.player
        extra_winners = {suit: [] for suit in SUITS}
        our_cards = player.our_unplayed_cards
        opponents_cards = player.opponents_unplayed_cards
        honours = 'AKQJT'
        for suit in SUITS:
            our_length = self._get_our_length(player, suit)
            # if suit == 'C' and (player.is_declarer or player.is_dummy):
            #     cprint(f'x {long_hand}', COLOUR)
            for honour in honours:
                cards = our_cards[suit]
                test_card = Card(honour, suit)
                if test_card in opponents_cards[suit]:
                    test_cards = []
                    for rank in honours.replace(honour, ''):
                        test_cards.append(Card(rank, suit))
                    for index, card in enumerate(test_cards):
                        if card in cards and our_length >= index + 2:
                            if (card not in extra_winners[suit] and
                                    card not in self.winners[suit]):
                                extra_winners[suit].append(card)
        return extra_winners

    def _get_winner_count(self) -> int:
        """Return the total number of winners."""
        winner_count = 0
        for suit, cards in self.winners.items():
            winner_count += len(cards)
        return winner_count



""" Third seat card play."""

from bridgeobjects import Card, Denomination
from bfgdealer import Trick

from .player import Player
from .utilities import trick_card_values, get_seat

PLAYER_INDEX = 2


class ThirdSeat():
    def __init__(self, player: Player):
        self.player = player
        self.trick = player.board.tricks[-1]
        self.seat = get_seat(self.trick, PLAYER_INDEX)
        self.cards = player.cards_for_trick_suit(self.trick)

    def _ace_is_deprecated(self, trick: Trick, card: Card) -> bool:
        """Return True if the ace is not to be played."""
        player = self.player
        if card.value != 13:
            return False

        king = Card(f'K{card.suit.name}')
        # print(player.card_has_been_played(king), trick.cards[1] == king)
        if self.player.card_has_been_played(king) or trick.cards[1] == king:
            return False

        if len(player.total_unplayed_cards[trick.suit]) < 9:
            return False

        return True

    def _trump_partners_card(self) -> bool:
        """Return True if third hand is to overtump partner's lead."""
        player = self.player
        trick = self.trick
        values = trick_card_values(trick, player.trump_suit)
        opponents_cards = player.opponents_unplayed_cards[trick.suit.name]
        if values[1] > values[0] and trick.cards[0].suit != player.trump_suit:
            return True
        for card in opponents_cards:
            if card.value > trick.cards[0].value:
                return True
        return False

    # def _best_suit(self, player: Player) -> Suit:
    #     """Select suit for signal."""
    #     # TODO handle no points and equal suits
    #     cards = player.hand_cards.list
    #     suit_points = get_suit_strength(cards)
    #     max_points = 0
    #     best_suit = None
    #     for suit in SUITS:
    #         hcp = suit_points[suit]
    #         if hcp > max_points:
    #             max_points = hcp
    #             best_suit = suit
    #     if not best_suit:
    #         return player.longest_suit
    #     return Suit(best_suit)



from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('fourth_seat_defender.pbn')


def test_play_partners_suit():
    """Ensure notdiscard a winner."""
    board = boards['0']
    assert next_card(board).name != 'QC'



"""Common functionality for cardplay."""

from bridgeobjects import Suit, Card, SEATS, SUITS, CARD_VALUES, RANKS
from bfgbidding import Hand
from bfgdealer import Trick

import bfgcardplay.src.global_variables as global_vars
from .data_classes import Cards, CardArray
from .data_classes import SuitCards
from .utilities import get_list_of_best_scores, get_suit_strength

global_vars.initialize()


class Player():
    """Object to represent the player: their hand and memory."""
    def __init__(self, board):
        self.board = board
        if not self.board.tricks:
            self.board.tricks.append(Trick())

        # Seats and roles
        self.seat = self._get_seat()
        self.seat_index = SEATS.index(self.seat)
        self.partner_seat = self._get_partners_seat(self.seat)
        self.declarer = board.contract.declarer
        self.declarer_index = SEATS.index(self.declarer)
        self.is_declarer = self.seat == self.declarer
        self.dummy_index = (self.declarer_index + 2) % 4
        self.dummy = SEATS[self.dummy_index]
        self.is_dummy = self.seat == self.dummy
        self.is_defender = not (self.is_declarer or self.is_dummy)
        (self.right_hand_seat,
         self.left_hand_seat) = self._get_opponents_seats()
        self.short_suit_seat = None

        # Hands suits and cards
        self.trump_suit = self._get_trump_suit()
        self.hands = board.hands
        self.hand = board.hands[self.seat]
        self.partners_hand = None
        self.declarers_hand = board.hands[self.declarer]
        self.dummys_hand = board.hands[self.dummy]

        # self.hand_cards is the Cards object for unplayed cards
        self.hand_cards = Cards(board.hands[self.seat].unplayed_cards)
        self.suit_cards = self.hand_cards.by_suit
        self.unplayed_cards = self.hand_cards.by_suit

        # self.longest_suit = self.hand_cards.longest_suit
        self.longest_suit = self._longest_suit()

        our_cards = Cards(
            board.hands[self.seat].cards + board.hands[self.partner_seat].cards
            )
        my_unplayed_cards = board.hands[self.seat].unplayed_cards
        partners_unplayed_cards = board.hands[self.partner_seat].unplayed_cards
        our_unplayed_cards = Cards(my_unplayed_cards + partners_unplayed_cards)
        self.our_cards: dict[str, list[Card]] = our_cards.by_suit
        self.our_unplayed_cards: dict[str, list[Card]] = {}
        self.our_unplayed_cards = our_unplayed_cards.by_suit

        (self.opponents_cards,
         self.opponents_unplayed_cards) = self._get_opponents_cards()

        self.seats_unplayed_cards = {
            seat: Cards(self.board.hands[seat].unplayed_cards).by_suit
            for seat in SEATS}
        self.highest_card = {seat: {} for seat in SEATS}

        # Trumps
        (self.trump_cards, self.opponents_trumps) = self._get_trump_cards()

        # Partners
        self.partners_hand = self._get_partners_hand()
        (self.partners_suit_cards,
         self.partners_unplayed_cards) = self._get_partners_suit_cards()
        self.total_unplayed_cards = self._total_unplayed_cards()

        # Declarer and Dummy
        # These card properties must be assigned in this order
        declarers_cards = Cards(board.hands[self.declarer].unplayed_cards)
        self.declarers_suit_cards = declarers_cards.by_suit
        dummys_cards = Cards(board.hands[self.dummy].unplayed_cards)
        self.dummys_suit_cards: dict[str, list[Card]] = dummys_cards.by_suit
        self.dummys_unplayed_cards: dict[str, list[Card]] = {}
        self.dummys_unplayed_cards = dummys_cards.by_suit
        self.dummy_suit_strength = get_suit_strength(dummys_cards.list)
        self.dummy_suit_tenaces = self._get_tenaces(self.dummys_suit_cards)

        # Short trump hands
        self.short_suit_seat = self._get_short_suit_seat()
        self.short_hand = self._get_short_hand()

        # Defenders
        (self.dummy_on_left, self.dummy_on_right) = self._get_dummy_location()

        # Information
        self.trick_number = len(board.tricks)
        self.declarers_target = self.board.contract.target_tricks
        (self.declarers_tricks,
         self.defenders_tricks) = self._get_trick_count()
        self.suit_rounds = self._get_suit_rounds()
        self.voids = self._get_voids()
        self.controls = self._get_controls()
        self.second_round_controls = self._get_second_round_controls()

        self.winners = self._get_suit_winners()
        self.threats = self._get_threats()
        self.winner_count = self._get_winner_count()
        self.losers = self._get_suit_losers()
        self.sure_tricks = CardArray(self._get_current_sure_tricks())
        self.tricks_needed = 6 + int(self.board.contract.name[0])
        self.probable_tricks = CardArray(self._get_current_probable_tricks())
        self.possible_tricks = CardArray(self._get_current_possible_tricks())
        self.possible_promotions = CardArray(self._get_possible_promotions())

    @staticmethod
    def cards_by_suit(hand: Hand) -> dict[str, list[Card]]:
        """Return a dict of unplayed cards by suit for the given hand."""
        cards = Cards(hand.unplayed_cards)
        return cards.by_suit

    @staticmethod
    def _get_partners_seat(seat: str) -> str:
        """Return partner's seat."""
        seat_index = SEATS.index(seat)
        partners_index = (seat_index + 2) % 4
        partners_seat = SEATS[partners_index]
        return partners_seat

    def _get_seat(self) -> str:
        """Return the current user's seat."""
        trick = self.board.tricks[-1]
        leader = trick.leader
        if trick.winner:
            leader = trick.winner
        leader_index = SEATS.index(leader)
        seat_index = (leader_index + len(trick.cards)) % 4
        seat = SEATS[seat_index]
        return seat

    def cards_for_trick_suit(self, trick: Trick) -> list[Card]:
        """Return a list of cards in the trick suit."""
        return self.hand_cards.by_suit[trick.suit]

    def _get_trump_suit(self) -> Suit:
        """Return the trump suit for the board (if any)."""
        if self.board.contract.is_nt:
            return None
        return self.board.contract.denomination

    def get_strongest_suits(self, cards: list[Card]) -> list[Suit]:
        """Return a list of suits that have the highest high card points."""
        suit_points = get_suit_strength(cards)
        strong_suits = get_list_of_best_scores(suit_points)
        return strong_suits

    def _get_tenaces(self, cards: dict[str, Card]) -> dict[str, Card]:
        """Return a dict of suit tenaces keyed on suit name."""
        suit_tenaces = {suit_name: None for suit_name, suit in SUITS.items()}
        for suit, cards in cards.items():
            suit_tenaces[suit] = self.get_suit_tenaces(cards)
        return suit_tenaces

    def get_suit_tenaces(self, cards: list[Card]) -> Card:
        """Return the top card in a tenaces, or None."""
        if not cards:
            return None

        opponents_cards = self.opponents_unplayed_cards[cards[0].suit.name]
        suit = cards[0].suit_name
        for card in cards[:-2]:
            if card.is_honour:
                value = card.value
                missing_card = Card(CARD_VALUES[value-1], suit)
                next_card = Card(CARD_VALUES[value-2], suit)
                if missing_card in opponents_cards and next_card in cards:
                    return next_card
        return None

    def _get_opponents_seats(self) -> tuple[str, str]:
        """Return a tuple with right and left hand seats."""
        seat_index = SEATS.index(self.seat)
        left_hand_seat_index = (seat_index + 1) % 4
        right_hand_seat_index = (seat_index + 3) % 4
        return (SEATS[right_hand_seat_index], SEATS[left_hand_seat_index])

    def _get_short_suit_seat(self):
        """Return the short seat in a suit contract."""
        if not self.trump_suit:
            return None
        declarers_trumps = self.declarers_suit_cards[self.trump_suit]
        dummys_trumps = self.dummys_suit_cards[self.trump_suit]
        if len(declarers_trumps) > len(dummys_trumps):
            return self.dummy
        return self.declarer

    def _get_trump_cards(self):
        """Return a tuple containing our and opponent's trump cards"""
        if self.trump_suit:
            trump_cards = self.hand_cards.by_suit[self.trump_suit]
            opponents_trumps = self.opponents_unplayed_cards[self.trump_suit]
            return (trump_cards, opponents_trumps)
        return (None, None)

    def _get_partners_hand(self):
        """Return partner's hand."""
        if self.is_declarer:
            return self.board.hands[self.dummy]
        elif self.is_dummy:
            return self.board.hands[self.declarer]
        return None

    def _get_partners_suit_cards(self):
        """Return partner's suit cards"""
        if not self.partners_hand:
            return ({suit: [] for suit in SUITS}, {suit: [] for suit in SUITS})
        partners_cards = Cards(self.partners_hand.cards)
        partners_unplayed_cards = Cards(self.partners_hand.unplayed_cards)
        return (partners_cards.by_suit, partners_unplayed_cards.by_suit)

    def _get_dummy_location(self):
        """Return a tuple with dummy's location."""
        if not self.is_defender:
            return (False, False)
        dummy_on_left = ((self.seat_index - self.dummy_index) % 4) == 3
        dummy_on_right = not dummy_on_left
        return (dummy_on_left, dummy_on_right)

    def _get_opponents_cards(self):
        """Return opponent's cards."""
        opponents_seats = (SEATS[(self.seat_index + 1) % 4],
                           SEATS[(self.seat_index + 3) % 4])
        opps_cards_0 = self.board.hands[opponents_seats[0]].cards
        opps_cards_1 = self.board.hands[opponents_seats[1]].cards
        opponents_cards = Cards(opps_cards_0 + opps_cards_1)
        unplayed_cards_0 = self.board.hands[opponents_seats[0]].unplayed_cards
        unplayed_cards_1 = self.board.hands[opponents_seats[1]].unplayed_cards
        opponents_unplayed_cards = Cards(unplayed_cards_0 +
                                         unplayed_cards_1)
        return (opponents_cards.by_suit, opponents_unplayed_cards.by_suit)

    def _get_short_hand(self):
        """Return the short hand in a suit contract."""
        if not self.trump_suit:
            return None
        declarers_trumps = self.get_unplayed_cards_by_suit(self.trump_suit,
                                                           self.declarer)
        dummys_trumps = self.get_unplayed_cards_by_suit(self.trump_suit,
                                                        self.dummy)
        declarer_longer_trumps = len(declarers_trumps) > len(dummys_trumps)
        if declarer_longer_trumps:
            return self.board.hands[self.dummy]
        return self.board.hands[self.declarer]

    def card_has_been_played(self, card: Card) -> bool:
        """Return True if the card has already been played."""
        for seat in SEATS:
            hand = self.board.hands[seat]
            if card in hand.unplayed_cards:
                return False
        return True

    def get_unplayed_cards_by_suit(self, suit: Suit, seat: str) -> list[Card]:
        """Return a list containing declarers and
        opponents unplayed cards in a suit."""
        hand = self.board.hands[seat]
        cards = hand.cards_by_suit[suit.name]
        unplayed_cards = [card for card in cards
                          if card in hand.unplayed_cards]
        return unplayed_cards

    def _total_unplayed_cards(self) -> dict[str, list[Card]]:
        """Return a dict containing all unplayed cards by suit."""
        unplayed_card_list = []
        for seat in SEATS:
            unplayed_card_list += self.hands[seat].unplayed_cards
            unplayed_cards = Cards(unplayed_card_list)
        return unplayed_cards.by_suit

    @staticmethod
    def _sort_cards(cards: list[Card]) -> list[Card]:
        """Return a sorted list of cards."""
        return sorted(cards, reverse=True)

    def long_seat(self, suit: str) -> str:
        my_cards = self.unplayed_cards[suit]
        partners_cards = self.partners_unplayed_cards[suit]
        if len(my_cards) > len(partners_cards):
            return self.seat
        else:
            return self.partner_seat

    def holds_all_winners_in_suit(self, suit: Suit,
                                  trick: Trick = None) -> bool:
        """Return True if the partnership holds all the winners in a suit."""
        unplayed_cards = self.opponents_unplayed_cards[suit.name]
        opponents_unplayed_cards = [card for card in unplayed_cards]
        our_cards = [card for card in self.our_unplayed_cards[suit.name]]
        if trick:
            for card in trick.cards:
                if card.suit == suit:
                    if card in self.opponents_cards[suit.name]:
                        opponents_unplayed_cards.append(card)
                    else:
                        our_cards.append(card)
        opponents_unplayed_cards = Cards(opponents_unplayed_cards).list

        # long_player = self.long_seat(suit.name)
        our_length = len(self.our_unplayed_cards[suit.name])
        opponents_length = len(opponents_unplayed_cards)

        for index, card in enumerate(our_cards):
            if index >= min(opponents_length, our_length):
                return True
            if card.value < opponents_unplayed_cards[0].value:
                return False
        else:
            return False

    def dummy_holds_adjacent_card(self, card: Card) -> bool:
        """Return True if the hand contains the adjacent card."""
        cards = self.dummys_suit_cards[card.suit.name]
        for other_card in cards:
            if (other_card.value == card.value + 1 or
                    other_card.value == card.value - 1):
                return True
        return False

    def partnership_long_suits(self, ignore_trumps=True):
        """Return a list of the longest partnership suits."""
        suits = {suit_name: len(self.our_unplayed_cards[suit_name])
                 for suit_name in SUITS}
        if self.trump_suit:
            if ignore_trumps:
                suits.pop(self.trump_suit.name)
        long_suits = get_list_of_best_scores(suits)
        return long_suits

    def can_lead_toward_tenace(self, long_suit: str) -> bool:
        """Return True if we can lead to higher honour in partner's hand."""
        if (self.partners_unplayed_cards[long_suit] and
                self.suit_cards[long_suit]):
            partners_best_card = self.partners_unplayed_cards[long_suit][0]
            my_best_card = self.unplayed_cards[long_suit][0]
            if partners_best_card.is_honour:
                if partners_best_card.value > my_best_card.value + 1:
                    return True
        return False

    def get_winners(self) -> int:
        """Return the current number of winners for declarer."""
        winners = 0
        for suit_name in SUITS:
            for card in self.our_unplayed_cards[suit_name]:
                if self.opponents_unplayed_cards[suit_name]:
                    unplayed_cards = self.opponents_unplayed_cards[suit_name]
                    opponents_top_card_value = unplayed_cards[0].value
                    if card.value > opponents_top_card_value:
                        winners += 1
                    else:
                        break
                else:
                    winners += 1
        return winners

    def _get_controls(self) -> dict[str, int]:
        """Return the current number of winners for declarer."""
        controls = {suit_name: 0 for suit_name in SUITS}
        for suit_name in SUITS:
            for card in self.our_unplayed_cards[suit_name]:
                unplayed_cards = self.opponents_unplayed_cards[suit_name]
                if unplayed_cards:
                    if unplayed_cards and card.value < unplayed_cards[0].value:
                        break
                    controls[suit_name] += 1
                else:
                    controls[suit_name] += 1
        return controls

    def _get_second_round_controls(self) -> dict[str, int]:
        """Return the current number of second round controls for declarer."""
        controls = {suit_name: False for suit_name in SUITS}
        for suit_name in SUITS:
            if len(self.our_unplayed_cards[suit_name]) >= 2:
                if len(self.opponents_unplayed_cards[suit_name]) <= 1:
                    controls[suit_name] = True
                    continue

                our_unplayed_cards = self.our_unplayed_cards[suit_name]
                opponents_cards = self.opponents_unplayed_cards[suit_name]
                if our_unplayed_cards[1].value > opponents_cards[1].value:
                    controls[suit_name] = True
        return controls

    def _get_trick_count(self) -> int:
        """Return the number of tricks won by declarer."""
        if self.declarer in 'NS':
            return (self.board.NS_tricks, self.board.EW_tricks)
        return (self.board.EW_tricks, self.board.NS_tricks)

    def get_entries(self, hand):
        """Return the controlling card in a suit."""
        entries = {suit_name: [] for suit_name in SUITS}
        unplayed_cards = Cards(hand.unplayed_cards)
        suit_cards = unplayed_cards.by_suit
        for suit_name in SUITS:
            cards = suit_cards[suit_name]
            for card in cards:
                if self._is_master(card, entries):
                    entries[suit_name].append(card)
                else:
                    break
        return entries

    def is_master_card(self, card: Card) -> bool:
        """Return True if card is a master."""
        unplayed_cards = self.total_unplayed_cards[card.suit]
        for unplayed_card in unplayed_cards:
            if unplayed_card.value > card.value:
                return False
        return True

    def _is_master(self, card: Card,
                   other_masters: dict[str, list[Card]]) -> bool:
        """Return True if card is a master."""
        unplayed_cards = self.total_unplayed_cards[card.suit]
        for master_card in other_masters[card.suit.name]:
            if master_card in unplayed_cards:
                unplayed_cards.remove(master_card)
        for unplayed_card in unplayed_cards:
            if unplayed_card.value > card.value:
                return False
        return True

    def is_winner_declarer(self, card: Card, trick: Trick = None) -> bool:
        """Return True is the card is a winner for declarer."""
        suit = card.suit.name
        opponents_unplayed_cards = [
            card for card in self.opponents_unplayed_cards[suit]
            ]
        if trick:
            if trick.leader in (self.seat, self.partner_seat):
                if len(trick.cards) > 1:
                    if trick.cards[1].suit.name == suit:
                        opponents_unplayed_cards.append(trick.cards[1])
                    fourth_player = SEATS[(SEATS.index(trick.leader) + 3) % 4]
                    if self.voids[fourth_player][trick.suit.name]:
                        opponents_unplayed_cards = []
        if not opponents_unplayed_cards:
            return True
        if card.value > sorted(opponents_unplayed_cards,
                               reverse=True)[0].value:
            return True
        return False

    def is_winner_defender(self, trick_card: Card,
                           trick: Trick = None) -> bool:
        """Return True is the card is a winner for defender."""
        card_value = trick_card.value
        if trick:
            if trick.cards[0].value > trick_card.value:
                return False
            if len(trick.cards) == 3:
                value_2 = trick.cards[2].value
                if self.trump_suit:
                    if trick.cards[2].suit == self.trump_suit:
                        value_2 += 13
                    if trick.card.suit == self.trump_suit:
                        card_value += 13
                        card_value = trick_card.value
                if value_2 > card_value:
                    return False
            suit = trick.suit
        else:
            suit = trick_card.suit

        # Get all unplayed cards in the suit
        unplayed_cards = [card
                          for card in self.total_unplayed_cards[suit.name]]

        # Remove dummy's card if it's already played
        if self.dummy_on_right and trick:
            for card in self.dummys_unplayed_cards[suit.name]:
                unplayed_cards.remove(card)

        # Remove my unplayed cards
        for card in self.unplayed_cards[suit.name]:
            unplayed_cards.remove(card)

        if not unplayed_cards:
            return True
        if card_value > unplayed_cards[0].value:
            return True
        return False

    def get_entries_in_other_suits(self, hand: Hand, suit: Suit):
        """Return the entries in a hand other than in the given suit."""
        entries = []
        for suit_name in SUITS:
            if suit_name != suit.name:
                cards = self.get_entries(hand)[suit_name]
                entries.extend(cards)
        return entries

    def missing_honours(self, suit: Suit) -> list[Card]:
        """Return a list of missing honours in the suit."""
        our_cards = self.our_unplayed_cards[suit]
        missing_honours = []
        for index, rank in enumerate('AKQJT'):
            card = Card(rank, suit.name)
            if card in self.total_unplayed_cards[suit]:
                if card not in our_cards:
                    missing_honours.append(card)
        return missing_honours

    def control_suit(self, suit: Suit) -> bool:
        """Return True if player totally controls suit."""
        our_cards = self.our_unplayed_cards[suit.name]
        opponents_cards = self.opponents_unplayed_cards[suit.name]
        if not opponents_cards:
            return True
        index = len(opponents_cards) - 1
        if index > len(our_cards) - 1:
            return False
        if opponents_cards[0].value > our_cards[index].value:
            return True
        return True

    def card_from_tenace(self) -> Card:
        """Return the bottom from a tenace in my hand with one threat."""
        trick = self.board.tricks[-1]
        cards = self.cards_for_trick_suit(trick)
        for index, card in enumerate(cards[1:]):
            if cards[index].value - card.value > 1:
                threats = 0
                for opps_card in self.opponents_unplayed_cards[trick.suit]:
                    if cards[index].value > opps_card.value > card.value:
                        threats += 1
                if threats <= 1:
                    return card
        return None

    def card_from_tenace_threat(self) -> tuple[Card, Card]:
        """Return the top and bottom from a tenace in my hand."""
        trick = self.board.tricks[-1]
        suit = trick.suit.name
        cards = self.cards_for_trick_suit(trick)
        higher_card = None
        threats = self._get_threats()
        if not threats[suit]:
            return (None, None)

        for threat in threats[suit]:
            if threat.rank != 'A':
                for value in range(threat.value+1, 14):
                    card = Card(CARD_VALUES[value], suit)
                    if card in cards:
                        higher_card = card
                        break
                    elif card in self.total_unplayed_cards[suit]:
                        return (None, None)
                if not higher_card:
                    return (None, None)

                for value in reversed(range(threat.value)):
                    card = Card(CARD_VALUES[value], suit)
                    if card in cards:
                        return (higher_card, card)
                    if card in self.total_unplayed_cards[suit]:
                        return (None, None)
        return (None, None)

    def touching_honours(self, cards: list[Card]) -> Card | None:
        """Return the top of touching honours if
        there is one or None otherwise."""
        for index, card in enumerate(cards[:-1]):
            if card.is_honour and card.value == cards[index+1].value + 1:
                return card
        return None

    def touching_honours_in_hand(self, hand: Hand, suit: str) -> Card | None:
        """Return the top of touching honours if
        there is one or None otherwise."""
        cards = Cards(hand.unplayed_cards).by_suit[suit]
        for index, card in enumerate(cards[:-1]):
            if card.is_honour and card.value == cards[index+1].value + 1:
                return card
        return None

    def _longest_suit(self) -> Suit:
        """Return the longest suit in the hand."""
        lengths = {suit: 0 for suit in SUITS}
        for suit in SUITS:
            lengths[suit] = len(self.unplayed_cards[suit])
        long_suits = get_list_of_best_scores(lengths)
        if len(long_suits) == 1:
            return Suit(long_suits[0])

        strength = {suit: 0 for suit in long_suits}
        for suit in long_suits:
            strength[suit] = get_suit_strength(self.unplayed_cards[suit])
        strong_suits = get_list_of_best_scores(lengths)
        if len(strong_suits) == 1:
            return Suit(strong_suits[0])
        return Suit(strong_suits[0])

    def _get_suit_rounds(self) -> dict[str, int]:
        """Return a dict of rounds played by suit."""
        rounds = {suit: 0 for suit in SUITS}
        for trick in self.board.tricks[:-1]:
            rounds[trick.suit.name] += 1
        return rounds

    def is_tenace(self, card_1: Card, card_2: Card, gap: int = 1) -> bool:
        """Return True if card_1 and card_2 form a tenace."""
        if card_1.suit != card_2.suit:
            return False
        suit = card_1.suit.name
        our_cards = self.our_unplayed_cards[suit]
        if card_1.value > card_2.value:
            top_card = card_1
            bottom_card = card_2
        elif card_1.value < card_2.value:
            top_card = card_2
            bottom_card = card_1
        else:
            return False

        cards = SuitCards(suit).sorted_cards()
        missing_masters = 0
        top_card_check = True
        bottom_card_check = True
        for card in cards:
            if (top_card_check and card != top_card and
                    card not in our_cards and
                    card in self.total_unplayed_cards[suit]):
                return False
            if bottom_card_check and not top_card_check:
                if (card != bottom_card and
                        card not in our_cards and
                        card in self.total_unplayed_cards[suit]):
                    missing_masters += 1
            if card == top_card:
                top_card_check = False
            if card == bottom_card:
                bottom_card_check = False
        if missing_masters == gap:
            return True
        return False

    def _get_threats(self) -> dict[str, Card]:
        """Return a dict of cards by suit that threaten players cards."""
        if self.trump_suit:
            return self._get_threats_suit_contract()
        else:
            return self._get_threats_nt_contract()

    def _get_threats_nt_contract(self) -> dict[str, Card]:
        """Return a dict of cards by suit that threaten players cards."""

        threats = {suit: [] for suit in SUITS}
        ranks = [rank for rank in RANKS]
        ranks.reverse()
        sorted_ranks = ranks[:-1]

        for suit in SUITS:
            long_cards = self.unplayed_cards[suit]
            our_cards = self.our_unplayed_cards[suit]
            opponents_cards = self.opponents_unplayed_cards[suit]
            missing = len(opponents_cards)
            winners = 0
            skip_card = False
            for index, rank in enumerate(sorted_ranks):
                ranking_card = Card(rank, suit)
                if ranking_card in self.total_unplayed_cards[suit]:
                    if ranking_card in our_cards:
                        if not skip_card:
                            winners += 1
                        skip_card = False

                        # No more cards to lose
                        if winners > missing:
                            break
                    else:
                        threats[suit].append(ranking_card)
                        # skip card means that the next card in the suit is
                        # assumed to have been beaten
                        skip_card = True

                    # threats only in long hand
                    if index + 1 >= len(long_cards):
                        break
                if len(opponents_cards) <= index + 1:
                    break
        return threats

    def _get_threats_suit_contract(self) -> dict[str, Card]:
        """Return a dict of cards by suit that threaten players cards."""
        (long_hand, short_hand) = self.get_long_short_trump_hands()
        unplayed_cards = self.cards_by_suit(long_hand)
        long_cards = unplayed_cards[self.trump_suit.name]

        threats = {suit: [] for suit in SUITS}
        ranks = [rank for rank in RANKS]
        ranks.reverse()
        sorted_ranks = ranks[:-1]

        for suit in SUITS:
            our_cards = self.our_unplayed_cards[suit]
            my_cards = self.unplayed_cards[suit]
            partners_cards = self.partners_unplayed_cards[suit]
            max_length = max(len(my_cards), len(partners_cards))
            opponents_cards = self.opponents_unplayed_cards[suit]
            missing = len(opponents_cards)
            winners = 0
            skip_card = False
            for index, rank in enumerate(sorted_ranks):
                ranking_card = Card(rank, suit)
                if ranking_card in self.total_unplayed_cards[suit]:
                    if ranking_card in our_cards:
                        if not skip_card:
                            winners += 1
                        skip_card = False

                        # Only count to the number of cards in our hands
                        if index + 1 >= max_length:
                            break

                        # No more cards to lose
                        if winners > missing:
                            break
                    else:
                        threats[suit].append(ranking_card)
                        # skip card means that the next card in the suit is
                        # assumed to have been beaten
                        skip_card = True

                    # threats only in long hand
                    if index + 1 >= len(long_cards):
                        break
                if len(opponents_cards) <= index + 1:
                    break
        return threats

    def get_long_short_trump_hands(self) -> tuple[Hand, Hand]:
        """Return the  hands for long and short trump hand between
        player and partner."""
        return self.get_long_short_hands(self.trump_suit)

    def get_long_short_hands(self, suit: Suit) -> tuple[Hand, Hand]:
        """Return the  hands for long and short hand between
        player and partner."""
        my_unplayed_cards = self.unplayed_cards[suit.name]
        partners_unplayed_cards = self.partners_unplayed_cards[suit.name]
        if len(my_unplayed_cards) >= len(partners_unplayed_cards):
            return (self.hand, self.partners_hand)
        else:
            return (self.partners_hand, self.hand)

    def _get_suit_winners(self) -> dict[str, Card]:
        """Return a dict of cards by suit that are certain winners."""
        winners = {suit: [] for suit in SUITS}
        our_cards = self.our_unplayed_cards
        opponents_cards = self.opponents_unplayed_cards
        for suit in SUITS:
            our_length = self._get_our_length(suit)
            for card in our_cards[suit]:
                if len(winners[suit]) < our_length:
                    if opponents_cards[suit]:
                        if card.value > opponents_cards[suit][0].value:
                            winners[suit].append(card)
                    else:
                        winners[suit].append(card)
        return winners

    def _get_suit_losers(self) -> dict[str, Card]:
        """Return a dict of cards by suit that are potential losers."""
        losers = {suit: [] for suit in SUITS}
        our_cards = self.our_unplayed_cards
        opponents_cards = self.opponents_unplayed_cards
        for suit in SUITS:
            our_length = self._get_our_length(suit)
            opponents_top_card = 0
            for card in our_cards[suit][:our_length]:
                if len(losers[suit]) < our_length:
                    if len(opponents_cards[suit]) >= opponents_top_card + 1:
                        opps_card = opponents_cards[suit][opponents_top_card]
                        top_value = opps_card.value
                        if card.value < top_value:
                            losers[suit].append(card)
                            opponents_top_card += 1
        return losers

    def _get_our_length(self, suit: str) -> int:
        my_cards = self.unplayed_cards
        partners_cards = self.partners_unplayed_cards
        if len(my_cards[suit]) > len(partners_cards[suit]):
            return len(my_cards[suit])
        return len(partners_cards[suit])

    def _get_current_sure_tricks(self) -> list[Card]:
        """Return a list of certain winners based on unplayed cards."""
        sure_tricks = []
        declarers_cards = self.unplayed_cards
        dummys_cards = self.partners_unplayed_cards
        for suit in SUITS:
            declarer_suit_cards = [card for card in declarers_cards[suit]]
            dummy_suit_cards = [card for card in dummys_cards[suit]]
            our_cards = CardArray.sort_cards(
                declarer_suit_cards + dummy_suit_cards
                )
            our_length = max(len(declarer_suit_cards), len(dummy_suit_cards))
            for card in our_cards[:our_length]:
                if self.is_winner_declarer(card, self.board.tricks[-1]):
                    sure_tricks.append(card)
        return sure_tricks

    def _get_current_probable_tricks(self) -> tuple[dict[str, list[Card]]]:
        """Return a dict of probable winners by suit based on
        unplayed cards."""
        opponents_length = {
            8: 5,
            7: 4,
            6: 4,
            5: 3,
            4: 3,
            3: 2,
            2: 2,
            1: 1,
            0: 0,
        }
        return self._current_trick_calculation(opponents_length)

    def _get_current_possible_tricks(self) -> tuple[dict[str, list[Card]]]:
        """Return a dict of possible winners by suit based on
        unplayed cards."""
        opponents_length = {
            8: 4,
            7: 4,
            6: 3,
            5: 3,
            4: 2,
            3: 2,
            2: 1,
            1: 1,
            0: 0,
        }
        return self._current_trick_calculation(opponents_length)

    def _current_trick_calculation(self,
                                   opponents_length: dict[int, int]
                                   ) -> tuple[dict[str, list[Card]]]:
        """Return a dict of winners by suit based on unplayed cards."""
        probable_tricks = []
        declarers_cards = self.unplayed_cards
        dummys_cards = self.partners_unplayed_cards
        opponents_cards = self.opponents_unplayed_cards
        for suit in SUITS:
            declarer_suit_cards = [card for card in declarers_cards[suit]]
            dummy_suit_cards = [card for card in dummys_cards[suit]]
            opponents_holding = (len(self.total_unplayed_cards[suit])
                                 - len(declarer_suit_cards)
                                 - len(dummy_suit_cards))

            if opponents_holding in opponents_length:
                # Honours
                honours = SuitCards(suit).cards('A', 'T')
                opponents_honours = 0
                for index, card in enumerate(honours):
                    if index < opponents_length[opponents_holding]:
                        if card in opponents_cards[suit]:
                            opponents_honours += 1
                opponents_count = min(
                    opponents_length[opponents_holding],
                    opponents_honours
                    )
                opponents_count = opponents_length[opponents_holding]

                # Length
                our_length = max(
                    len(declarer_suit_cards),
                    len(dummy_suit_cards)
                    )
                if len(declarer_suit_cards) > len(dummy_suit_cards):
                    our_long_hand = declarer_suit_cards
                else:
                    our_long_hand = dummy_suit_cards

                if our_long_hand:
                    range_limit = our_length - opponents_count
                    for index in range(range_limit):
                        card = our_long_hand[-index]
                        if card not in self.sure_tricks.cards:
                            probable_tricks.append(card)
        return probable_tricks

    def _get_possible_promotions(self):
        """Return a dict of possible promotions
        by suit based on unplayed cards."""
        def check_card(self, hand, card, buffer):
            """Return True if the card is a candidate for promotion."""
            if (card in hand and
                    len(hand) >= buffer and
                    card not in self.sure_tricks.cards):
                return True
            return False

        possible_promotions = []
        declarers_cards = self.unplayed_cards
        dummys_cards = self.partners_unplayed_cards
        for suit in SUITS:
            if suit not in self.probable_tricks.suits:
                honours = SuitCards(suit).cards('K', 'T')
                for hand in [declarers_cards[suit], dummys_cards[suit]]:
                    for index, card in enumerate(honours):
                        if check_card(self, hand, card, 2+index):
                            possible_promotions.append(card)
        return possible_promotions

    def _get_current_entries(self) -> tuple[dict[str, list[Card]]]:
        """Return a dict of probable entries by
        suit based on unplayed cards."""
        my_entries = {suit: [] for suit in SUITS}
        partners_entries = {suit: [] for suit in SUITS}
        for suit in SUITS:
            my_cards = self.unplayed_cards[suit]
            partners_cards = self.partners_unplayed_cards[suit]
            if my_cards and partners_cards:
                if (self.is_winner_declarer(my_cards[0]) and
                        my_cards[0].value > partners_cards[-1].value):
                    my_entries[suit] = my_cards[0]
                if (self.is_winner_declarer(partners_cards[0]) and
                        partners_cards[0].value > my_cards[-1].value):
                    partners_entries[suit] = partners_cards[0]
        return (my_entries, partners_entries)

    def _get_voids(self):
        # Record voids
        voids = {seat: {suit: False for suit in SUITS} for seat in SEATS}
        for trick in self.board.tricks:
            for index, card in enumerate(trick.cards[1:]):
                if card.suit != trick.suit:
                    seat_index = (SEATS.index(trick.leader) + index + 1) % 4
                    voids[SEATS[seat_index]][trick.suit.name] = True
        return voids

    def _get_winner_count(self) -> int:
        """Return the total number of winners."""
        winner_count = 0
        for cards in self.winners.values():
            winner_count += len(cards)
        return winner_count

    # def _get_threats(self) -> dict[str, Card]:
    #     """Return a dict of cards by suit that threaten players cards."""
    #     if self.trump_suit:
    #         return self._get_threats_suit_contract()
    #     return self._get_threats_nt_contract()

    # def _get_threats_nt_contract(self) -> dict[str, Card]:
    #     """Return a dict of cards by suit that threaten players cards."""
    #     threats = {suit: [] for suit in SUITS}
    #     ranks = [rank for rank in RANKS]
    #     ranks.reverse()
    #     sorted_ranks = ranks[:-1]

    #     for suit in SUITS:
    #         long_cards = self.unplayed_cards[suit]
    #         our_cards = self.our_unplayed_cards[suit]
    #         opponents_cards = self.opponents_unplayed_cards[suit]
    #         missing = len(opponents_cards)
    #         winners = 0
    #         skip_card = False
    #         for index, rank in enumerate(sorted_ranks):
    #             ranking_card = Card(rank, suit)
    #             if ranking_card in self.total_unplayed_cards[suit]:
    #                 if ranking_card in our_cards:
    #                     if not skip_card:
    #                         winners += 1
    #                     skip_card = False

    #                     # No more cards to lose
    #                     if winners > missing:
    #                         break
    #                 else:
    #                     threats[suit].append(ranking_card)
    #                     # skip card means that the next card in the suit is
    #                     # assumed to have been beaten
    #                     skip_card = True

    #                 # threats only in long hand
    #                 if index + 1 >= len(long_cards):
    #                     break
    #             if len(opponents_cards) <= index + 1:
    #                 break
    #     return threats

    # def _get_threats_suit_contract(self) -> dict[str, Card]:
    #     """Return a dict of cards by suit that threaten players cards."""
    #     (long_hand, short_hand) = self.get_long_short_trump_hands()
    #     unplayed_cards = self.cards_by_suit(long_hand)
    #     long_cards = unplayed_cards[self.trump_suit.name]

    #     threats = {suit: [] for suit in SUITS}
    #     ranks = [rank for rank in RANKS]
    #     ranks.reverse()
    #     sorted_ranks = ranks[:-1]

    #     for suit in SUITS:
    #         our_cards = self.our_unplayed_cards[suit]
    #         my_cards = self.unplayed_cards[suit]
    #         partners_cards = self.partners_unplayed_cards[suit]
    #         max_length = max(len(my_cards), len(partners_cards))
    #         opponents_cards = self.opponents_unplayed_cards[suit]
    #         missing = len(opponents_cards)
    #         winners = 0
    #         skip_card = False
    #         for index, rank in enumerate(sorted_ranks):
    #             ranking_card = Card(rank, suit)
    #             if ranking_card in self.total_unplayed_cards[suit]:
    #                 if ranking_card in our_cards:
    #                     if not skip_card:
    #                         winners += 1
    #                     skip_card = False

    #                     # Only count to the number of cards in our hands
    #                     if index + 1 >= max_length:
    #                         break

    #                     # No more cards to lose
    #                     if winners > missing:
    #                         break
    #                 else:
    #                     threats[suit].append(ranking_card)
    #                     # skip card means that the next card in the suit is
    #                     # assumed to have been beaten
    #                     skip_card = True

    #                 # threats only in long hand
    #                 if index + 1 >= len(long_cards):
    #                     break
    #             if len(opponents_cards) <= index + 1:
    #                 break
    #     return threats



"""Second seat card play for defender."""
from termcolor import colored

import inspect

from ..logger import log

from bridgeobjects import Card, CARD_VALUES, SUITS, Suit
import bfgcardplay.src.global_variables as global_vars
from .player import Player
from .utilities import get_suit_strength
from .second_seat import SecondSeat
from .defender_play import get_hilo_signal_card, signal_card, surplus_card, best_discard
from .data_classes import SuitCards

MODULE_COLOUR = 'green'


class SecondSeatDefender(SecondSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the card if the second seat."""
        player = self.player
        trick = self.trick
        cards = self.cards

        # Singleton
        if len(cards) == 1:
            return log(inspect.stack(), cards[0])

        # Void
        if not cards:
            return self._select_card_if_void()

        if (player.is_winner_defender(cards[0], trick) and
                len(player.dummys_unplayed_cards[trick.suit.name]) <= 2):
            return log(inspect.stack(), cards[0])

        # Cover honour with honour
        card = self._cover_honour_with_honour()
        if card:
            return card

        # If winner and last opportunity to play it
        card = self._play_winner_if_last_opportunity()
        if card:
            return card

        # Play honour if higher honour in dummy
        card = self._beat_dummys_honour()
        if card:
            return card

        # Win trick if possible
        card = self._play_winner()
        if card:
            return card

        # Play K if doubleton
        card = self._play_doubleton_king()
        if card:
            return card

        # Signal even/odd
        card = self._signal_even_odd()
        if card:
            return card

        # Play lowest card
        return log(inspect.stack(), cards[-1])

    def _play_winner_if_last_opportunity(self) -> Card:
        player = self.player
        cards = self.cards
        trick = self.trick
        if player.trump_suit and cards:
            opponents_cards = player.opponents_unplayed_cards[trick.suit.name]
            if opponents_cards:
                if player.is_winner_defender(cards[0], trick):
                    return log(inspect.stack(), cards[0])

        # else:  # TODO add something for NT contracts
        #         for card in cards[::-1]:
        #             if card.value > trick.cards[0].value:
        #                 return card
        return None

    def _play_winner(self) -> Card:
        player = self.player
        cards = self.cards
        suit = self.trick.suit.name
        if self.trick.cards[0].value > player.unplayed_cards[suit][0].value:
            return None

        if player.trick_number >= 9:
            if not player.dummys_unplayed_cards[suit]:
                dummys_top_value = 0
            else:
                dummys_top_value = player.dummys_unplayed_cards[suit][0].value

            dummy_safe = ((player.dummy_on_right and
                          cards[0].value < dummys_top_value) or
                          (player.dummy_on_left and
                          cards[0].value > dummys_top_value))
            if dummy_safe:
                return log(inspect.stack(), cards[0])
        return None

    def _play_doubleton_king(self) -> Card:
        player = self.player
        suit = self.trick.suit.name
        (ace, king) = SuitCards(suit).cards('A', 'K')
        ace_safe = (player.dummy_on_right and ace in player.dummys_unplayed_cards[suit] or
                    player.dummy_on_left and ace not in player.dummys_unplayed_cards[suit])
        if len(player.unplayed_cards) == 2 and self.cards[0] == king and ace_safe:
            return log(inspect.stack(), king)
        return None

    def _signal_even_odd(self) -> Card:
        player = self.player
        if not self.trick.suit == player.trump_suit:
            if len(player.unplayed_cards) > 2 and not self.cards[0].is_honour:
                card = get_hilo_signal_card(self)
                if card:
                    return card
        return None

    def _cover_honour_with_honour(self) -> Card:
        # TODO see this web site http://www.rpbridge.net/4l00.htm
        player = self.player
        cards = self.cards
        trick = self.trick
        cover_allowed = True
        if player.dummy_on_right:
            if player.dummy_holds_adjacent_card(trick.cards[0]):
                cover_allowed = False

        if cover_allowed and trick.cards[0].value >= CARD_VALUES['9']:  # nine or above
            if len(cards) >= 2:
                if cards[1].value >= CARD_VALUES['T']:
                    for card in cards[::-1]:
                        if card.value > trick.cards[0].value:
                            return log(inspect.stack(), card)
        return None

    def _beat_dummys_honour(self) -> Card:
        player = self.player
        suit_name = self.trick.suit.name
        if player.dummy_on_right:
            for card in self.cards:
                if card.is_honour and card.rank != 'A':
                    value = card.value
                    test_card = Card(CARD_VALUES[value+1], suit_name)
                    if test_card in player.dummys_unplayed_cards[suit_name]:
                        return log(inspect.stack(), card)
        return None

    def _select_card_if_void(self) -> Card:
        """Return card if cannot follow suit."""
        player = self.player
        trick = self.trick
        manager = global_vars.manager
        if player.trump_suit and player.unplayed_cards[player.trump_suit] and player.is_defender:
            return log(inspect.stack(), player.unplayed_cards[player.trump_suit][-1])

        best_suit = self._best_suit(player)

        # Signal suit preference."""
        card = signal_card(player, manager, best_suit)
        if card:
            return log(inspect.stack(), card)

        # Discard if more cards than the opposition
        card = surplus_card(player)
        if card:
            return log(inspect.stack(), card)

        # Best discard
        return best_discard(player)

    def _best_suit(self, player: Player) -> Suit:
        """Select suit for signal."""
        # TODO handle no points and equal suits
        cards = player.hand_cards.list
        suit_points = get_suit_strength(cards)
        max_points = 0
        best_suit = None
        for suit in SUITS:
            hcp = suit_points[suit]
            if hcp > max_points:
                max_points = hcp
                best_suit = suit
        if not best_suit:
            return player.longest_suit
        return Suit(best_suit)



from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('second_seat_declarer.pbn')


def test_play_partners_suit():
    """Do not play high card when inappropriate."""
    board = boards['0']
    assert next_card(board).name == '7H'


def test_do_not_signal_with_honour():
    """Do not signal with a losing honour."""
    board = boards['1']
    assert next_card(board).name != 'KS'

def test_do_play_high_in_second_seat():
    """Do play high in second seat in NT."""
    board = boards['2']
    assert next_card(board).name != 'KD'



from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('third_seat_declarer.pbn')


def test_play_partners_suit():
    "Do not play QC on AC."""
    board = boards['0']
    assert next_card(board).name == '5C'


def test_play_a_heart_when_all_winners():
    board = boards['1']
    assert next_card(board).suit.name != 'H'


def test_detect_void_in_fourth_hand():
    board = boards['2']
    assert next_card(board).name == 'TD'



""" First seat card play for defender."""

from termcolor import cprint, colored

import inspect
from ..logger import log

from bridgeobjects import SUITS, Card, Suit, CARD_VALUES
from .player import Player
from .first_seat import FirstSeat
from .defender_play import deduce_partner_void_in_trumps, dummy_is_short_trump_hand
import bfgcardplay.src.global_variables as global_vars

MODULE_COLOUR = 'blue'


class FirstSeatDefender(FirstSeat):
    def __init__(self, player: Player):
        super().__init__(player)

    def selected_card(self) -> Card:
        """Return the lead card for first seat defender."""
        player = self.player
        self.update_manager()
        if len(player.hand.unplayed_cards) == 1:
            return log(inspect.stack(), player.hand.unplayed_cards[0])

        if player.board.contract.is_nt:
            suit = self._select_suit_for_nt_contract()
        else:
            suit = self._select_suit_for_suit_contract()
        if not player.unplayed_cards[suit]:
            raise ValueError(colored(f'No cards for {suit}', MODULE_COLOUR))
        card = self._select_card_from_suit(suit)
        return card

    def _select_suit_for_suit_contract(self) -> Suit:
        """Return the trick lead suit for the defending a suit contract."""
        player = self.player
        manager = global_vars.manager

        suit = self._play_partners_suit()
        if suit:
            return suit

        suit = self._partner_void_in_trumps()
        if suit:
            return suit

        # Play winner if a singleton in dummy
        suit = self._singleton_winner_in_dummy()
        if suit:
            return suit

        # Select suit if partner likes it
        for suit in SUITS:
            if (manager.like_dislike(player.partner_seat, suit) and
                    player.unplayed_cards[suit] and
                    player.dummys_suit_cards[suit]):
                return log(inspect.stack(), Suit(suit))

        # Select a suit if not deprecated
        deprecated_suits = self._get_deprecated_suits()
        suit = self._select_non_deprecated_suit(deprecated_suits)
        if suit:
            return suit

        # Play through a tenace
        tenace_suits = self._get_tenace_suits(deprecated_suits)
        if tenace_suits:
            return log(inspect.stack(), tenace_suits[0])

        # Play a 'liked' suit
        suit = self._select_liked_suit()
        if suit:
            return suit

        # Lead through strength
        suit = self._play_through_strength()
        if suit:
            return suit

        # Dummy out of trumps
        suit = self._dummy_out_of_trumps()
        if suit:
            return suit

        score_reasons = self._score_reasons()
        # Select best suit
        best_suit = self._best_suit(score_reasons)
        return log(inspect.stack(), best_suit)

    def _play_partners_suit(self) -> Suit:
        player = self.player
        partners_suit = player.board.tricks[0].cards[0].suit.name
        if player.dummys_unplayed_cards[partners_suit]:
            if player.unplayed_cards[partners_suit]:
                return log(inspect.stack(), Suit(partners_suit))
        return None

    def _singleton_winner_in_dummy(self) -> Suit:
        player = self.player
        if player.trump_suit:
            for suit in SUITS:
                if suit != player.trump_suit and player.unplayed_cards[suit]:
                    if (len(player.dummys_unplayed_cards[suit]) == 1 and
                            player.is_winner_defender(player.unplayed_cards[suit][0])):
                        return log(inspect.stack(), Suit(suit))
        return None

    def _get_deprecated_suits(self) -> list[Suit]:
        deprecated_suits = []
        dummy_tenaces = self.identify_dummy_tenaces()
        for suit in SUITS:
            if self.player.dummy_on_right and suit in dummy_tenaces:
                deprecated_suits.append(suit)
        return deprecated_suits

    def _get_tenace_suits(self, deprecated_suits) -> list[Suit]:
        tenace_suits = []
        dummy_tenaces = self.identify_dummy_tenaces()
        for suit in SUITS:
            if suit not in deprecated_suits:
                if self.player.dummy_on_left and suit in dummy_tenaces:
                    tenace_suits.append(Suit(suit))
        return tenace_suits

    def _select_non_deprecated_suit(self, deprecated_suits) -> Suit:
        player = self.player
        for suit in SUITS:
            if suit not in deprecated_suits:
                dummy_short_trumps = dummy_is_short_trump_hand(player)
                if ((dummy_short_trumps and player.dummys_suit_cards[suit]) or
                        not dummy_short_trumps):
                    cards = player.unplayed_cards[suit]
                    if cards:
                        top_card = cards[0]

                        # Play next of touching
                        if top_card.value < CARD_VALUES['A']:
                            higher_rank = CARD_VALUES[top_card.value+1]
                            higher_card = Card(higher_rank, suit)
                            if top_card.is_honour and higher_card in player.hand.cards:
                                return log(inspect.stack(), Suit(suit))
                        # Play top of touching
                        if len(cards) > 1:
                            if top_card.is_honour and top_card.value == cards[1].value + 1:
                                return log(inspect.stack(), Suit(suit))
        return None

    def _select_liked_suit(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        for suit in SUITS:
            if (player.dummys_unplayed_cards[suit] or
                    not player.dummys_unplayed_cards[player.trump_suit]):
                if manager.like_dislike(player.partner_seat, suit) and player.unplayed_cards[suit]:
                    return log(inspect.stack(), Suit(suit))
                if manager.like_dislike(player.seat, suit) and player.unplayed_cards[suit]:
                    return log(inspect.stack(), Suit(suit))
        return None

    def _dummy_out_of_trumps(self) -> Suit:
        player = self.player
        if player.trump_suit:
            for suit in SUITS:
                if (len(player.total_unplayed_cards[suit]) == 1 and
                        player.total_unplayed_cards[suit][0] in player.unplayed_cards[suit] and
                        not player.dummys_suit_cards[player.trump_suit.name]):
                    return log(inspect.stack(), Suit(suit))
        return None

    def _score_reasons(self) -> dict[str, int]:
        score_reasons = {}

        # Deprecate voids
        score_reasons['void'] = self._deprecate_suits()

        # Trumps
        score_reasons['trumps'] = self._trumps()

        # Return partner's suit
        score_reasons['partner'] = self._partners_suit()

        # Lead from sequence
        score_reasons['sequences'] = self._sequences()

        # Lead to partner's void
        score_reasons['sequences'] = self._partners_voids()

        # Lead through tenaces not to tenaces
        score_reasons['tenaces'] = self._tenace_check()

        # Lead through or to strength
        score_reasons['weakness'] = self._lead_through_strength()

        # Avoid frozen suits
        score_reasons['frozen'] = self._frozen_suits()

        # Long suits
        score_reasons['long'] = self._long_suits()

        # Short suits
        score_reasons['short'] = self._short_suits()

        # Ruff and discard
        if self.player.trump_suit:
            score_reasons['ruff'] = self._ruff_and_discard()
        return score_reasons

    def _play_through_strength(self) -> Suit:
        player = self.player
        for suit in SUITS:
            if player.unplayed_cards[suit]:
                if player.dummy_on_left:
                    if player.dummys_unplayed_cards[suit]:
                        dummys_top_card = player.dummys_unplayed_cards[suit][0]
                        if (dummys_top_card.is_honour and
                                dummys_top_card.value < player.total_unplayed_cards[suit][0].value):
                            return log(inspect.stack(), Suit(suit))
        return None

    def _partner_void_in_trumps(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        # if deduce_partner_void_in_trumps(player):
        #     manager.voids[player.partner_seat][player.trump_suit.name] = True

        partners_seat = player.partner_seat
        partners_voids = player.voids[partners_seat]
        if not partners_voids[player.trump_suit.name]:
            for suit in SUITS:
                if (player.voids[partners_seat][suit] and
                        player.unplayed_cards[suit]):
                    manager.set_suit_to_develop(player.seat, Suit(suit))
                    manager.set_suit_strategy(
                        player.seat,
                        suit,
                        'ruff_in_void'
                    )
                    return log(inspect.stack(), Suit(suit))
        return None

    def _select_suit_for_nt_contract(self) -> Suit:
        """Return the trick lead suit for the defending a suit contract."""
        player = self.player

        # Return partner's suit
        if len(player.unplayed_cards[player.board.tricks[0].suit]) > 0:
            return log(inspect.stack(), player.board.tricks[0].suit)

        # Select suit after partner signal
        suit = self._suit_after_partner_signal()
        if suit:
            return suit

        suit = self._suit_to_develop()
        if suit:
            return suit

        suit = self._working_suit()
        if suit:
            return suit

        suit = self._set_suit_to_develop()
        if suit:
            return suit

        suit = self._is_winner_defender()
        if suit:
            return suit

        suit = self._liked_suit()
        if suit:
            return suit

        score_reasons = self._score_reasons()
        best_suit = self._best_suit(score_reasons)
        return log(inspect.stack(), best_suit)

    def _suit_after_partner_signal(self) -> Suit:
        player = self.player
        if len(player.board.tricks) == 2 and player.dummy_on_right:
            partners_lead = player.board.tricks[0].cards[0]
            if (partners_lead.value >= CARD_VALUES['T'] or
                    partners_lead.value < CARD_VALUES['6']):
                if player.unplayed_cards[partners_lead.suit.name]:
                    return log(inspect.stack(), partners_lead.suit)
        return None

    def _suit_to_develop(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        suit_to_develop = manager.suit_to_develop(player.seat)
        if suit_to_develop and player.unplayed_cards[suit_to_develop.name]:
            return log(inspect.stack(), suit_to_develop)
        return None

    def _working_suit(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        working_suit = manager.working_suit[player.seat]
        if working_suit and working_suit.name:
            if player.suit_cards[working_suit.name]:
                return log(inspect.stack(), working_suit)
        return None

    def _set_suit_to_develop(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        my_hand = player.hand
        longest_suit = my_hand.longest_suit
        if (my_hand.cards_by_suit[longest_suit.name][0].is_honour and
                player.get_entries_in_other_suits(my_hand, longest_suit) and
                player.unplayed_cards[longest_suit.name]):
            manager.set_suit_to_develop(player.seat, longest_suit)
            return log(inspect.stack(), longest_suit)
        return None

    def _is_winner_defender(self) -> Suit:
        player = self.player
        if len(player.board.tricks) > 6:
            for suit in SUITS:
                cards = player.unplayed_cards[suit]
                if cards:
                    if player.is_winner_defender(
                            player.unplayed_cards[suit][0]
                        ):
                        return log(inspect.stack(), Suit(suit))
        return None

    def _liked_suit(self) -> Suit:
        player = self.player
        manager = global_vars.manager
        for suit in SUITS:
            if (manager.like_dislike(player.partner_seat, suit) and
                    player.unplayed_cards[suit]):
                return log(inspect.stack(), Suit(suit))
        return None

    def _select_card_from_suit(self, suit: Suit) -> Card:
        """Return the card to lead from the given suit."""
        player = self.player
        manager = global_vars.manager
        cards = player.suit_cards[suit.name]

        suit_to_develop = manager.suit_to_develop(player.seat)
        if suit_to_develop:
            suit_cards = player.unplayed_cards[suit_to_develop]
            if suit_cards and len(player.total_unplayed_cards[suit_to_develop]) > 1:
                if manager.suit_strategy(player.seat)[suit_to_develop.name] == 'ruff_in_void':
                    if suit_cards[0].value > player.total_unplayed_cards[suit_to_develop][1].value:
                        return log(inspect.stack(), suit_cards[0])
                    else:
                        return log(inspect.stack(), suit_cards[-1])

        # Winning card
        if player.trump_suit:
            if player.is_winner_defender(player.unplayed_cards[suit][0]):
                return log(inspect.stack(), cards[0])

        if self._all_winners(suit):
            return log(inspect.stack(), player.unplayed_cards[suit][0])

        # Top of touching honours
        for index, card in enumerate(cards[:-1]):
            if card.is_honour and card.value == cards[index+1].value + 1:
                return log(inspect.stack(), card)

        # Top of doubleton
        if len(cards) == 2:
            return log(inspect.stack(), cards[0])

        # Return winners
        # TODO sort out invalid for defender
        winners = player.winners[suit.name]
        ic(winners)
        if winners and winners[-1] in cards:
            return log(inspect.stack(), winners[-1])

        # Return bottom card
        return log(inspect.stack(), cards[-1])

    def _sequences(self) -> Suit:
        """Return the score for sequences."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        player = self.player
        touching_honours = player.hand.touching_honours()
        best_sequence = None
        max_score = 0
        for suit_name in suit_scores:
            if touching_honours[suit_name]:
                suit_scores[suit_name] += self.TOUCHING_HONOURS
                suit_scores[suit_name] += len(touching_honours[suit_name])
                if suit_scores[suit_name] > max_score:
                    max_score = suit_scores[suit_name]
                    best_sequence = suit_name

        manager = global_vars.manager
        manager.working_suit[player.seat] = Suit(best_sequence)
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _partners_voids(self) -> Suit:
        """Return the score for sequences."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        player = self.player
        partners_voids = player.voids[player.seat]
        for suit, void in partners_voids.items():
            if void:
                suit_scores[suit] += self.PARTNERS_VOID
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _deprecate_suits(self) -> list[tuple[str, int]]:
        """Assign score to suits based on void."""
        player = self.player
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        for suit in SUITS:
            if not self.player.suit_cards[suit]:
                suit_scores[suit] -= self.VOID_SCORE
            if player.trump_suit:
                dummy_short_trumps = dummy_is_short_trump_hand(player)
                if (dummy_short_trumps and not player.dummys_suit_cards[suit]):
                    suit_scores[suit] -= self.RUFF_AND_DISCARD_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _trumps(self) -> list[tuple[str, int]]:
        """Assign score to suits based on void."""
        suit_scores = {suit_name: 0 for suit_name in SUITS}
        player = self.player
        if player.trump_suit:
            my_cards = player.unplayed_cards[player.trump_suit.name]
            dummys_cards = player.dummys_suit_cards[player.trump_suit.name]
            if len(my_cards) > len(dummys_cards):
                suit_scores[player.trump_suit.name] += self.TRUMP_SCORE
        return [(suit_name, score) for suit_name, score in suit_scores.items()]

    def _all_winners(self, suit: Suit) -> bool:
        """Return True if all of the cards held are winners."""
        player = self.player
        all_winners = True
        unplayed_cards = len(player.total_unplayed_cards[suit])
        missing_cards = unplayed_cards - len(player.suit_cards[suit])
        if len(player.unplayed_cards[suit]) < missing_cards:
            return False
        for index in range(missing_cards):
            if not player.is_winner_defender(
                    player.unplayed_cards[suit][index]
                ):
                all_winners = False
        return all_winners



from pathlib import Path
import logging

# Create a custom logger
logger = logging.getLogger(__name__)
logger.propagate = False
logging.basicConfig(level=logging.DEBUG)

# Create handlers
c_handler = logging.StreamHandler()
f_handler = logging.FileHandler('file.log')
c_handler.setLevel(logging.DEBUG)
f_handler.setLevel(logging.DEBUG)

# Create formatters and add it to handlers
c_format = logging.Formatter('bfgcp: %(message)s')
f_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
c_handler.setFormatter(c_format)
f_handler.setFormatter(f_format)

# Add handlers to the logger
logger.addHandler(c_handler)
logger.addHandler(f_handler)

def log(frame, obj):
    """Create log message and return the object."""
    path = Path(frame[0][1])
    module = path.stem
    line = frame[0][2]
    message_text = f'{obj} <{module}> {line}'
    logger.debug(message_text)
    return obj




from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('third_seat_defender.pbn')


def test_play_partners_suit():
    """Return win if possible in _select_card_based_on_position."""
    board = boards['0']
    assert next_card(board).name == 'AH'


def test_play_high_card():
    """Return highest card in _play_high_card."""
    board = boards['1']
    assert next_card(board).name == 'TS'


def test_dont_play_high_card_if_partner_played_higher():
    """Do Not return highest card in _play_high_card."""
    board = boards['2']
    assert next_card(board).name != 'KS'


def test_dont_play_high_card_if_partner_played_winner():
    board = boards['4']
    assert next_card(board).name != 'AC'


def test_all_winners_in_long_suit():
    # Play winner if all winners
    board = boards['3']
    assert next_card(board).name == 'KH'


def test_lead_unblock():
    # Play winner if all winners
    board = boards['5']
    assert next_card(board).name == '9C'
